"""Utilities for loading and dumping DAGs from/to YAML files."""
import copy
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from sky import dag as dag_lib
from sky import sky_logging
from sky import task as task_lib
from sky.skylet import constants
from sky.utils import cluster_utils
from sky.utils import registry
from sky.utils import ux_utils
from sky.utils import yaml_utils

logger = sky_logging.init_logger(__name__)

# JobGroup header fields
_JOB_GROUP_HEADER_FIELDS = {
    'name', 'execution', 'primary_tasks', 'termination_delay'
}
_JOB_GROUP_REQUIRED_HEADER_FIELDS = {'name'}

# Message thrown when APIs sky.{exec,launch,jobs.launch}() received a string
# instead of a Dag.  CLI (cli.py) is implemented by us so should not trigger
# this.
_ENTRYPOINT_STRING_AS_DAG_MESSAGE = """\
Expected a sky.Task or sky.Dag but received a string.

If you meant to run a command, make it a Task's run command:

    task = sky.Task(run=command)

The command can then be run as:

  sky.exec(task, cluster_name=..., ...)
  # Or use {'V100': 1}, 'V100:0.5', etc.
  task.set_resources(sky.Resources(accelerators='V100:1'))
  sky.exec(task, cluster_name=..., ...)

  sky.launch(task, ...)

  sky.jobs.launch(task, ...)
""".strip()


def convert_entrypoint_to_dag(entrypoint: Any) -> 'dag_lib.Dag':
    """Converts the entrypoint to a sky.Dag and applies the policy.

    Raises TypeError if 'entrypoint' is not a 'sky.Task' or 'sky.Dag'.
    """
    # Not suppressing stacktrace: when calling this via API user may want to
    # see their own program in the stacktrace. Our CLI impl would not trigger
    # these errors.
    converted_dag: 'dag_lib.Dag'
    if isinstance(entrypoint, str):
        with ux_utils.print_exception_no_traceback():
            raise TypeError(_ENTRYPOINT_STRING_AS_DAG_MESSAGE)
    elif isinstance(entrypoint, dag_lib.Dag):
        converted_dag = copy.deepcopy(entrypoint)
    elif isinstance(entrypoint, task_lib.Task):
        entrypoint = copy.deepcopy(entrypoint)
        with dag_lib.Dag() as dag:
            dag.add(entrypoint)
            dag.name = entrypoint.name
        converted_dag = dag
    else:
        with ux_utils.print_exception_no_traceback():
            raise TypeError(
                'Expected a sky.Task or sky.Dag but received argument of type: '
                f'{type(entrypoint)}')

    return converted_dag


def _load_chain_dag(
        configs: List[Dict[str, Any]],
        env_overrides: Optional[List[Tuple[str, str]]] = None,
        secrets_overrides: Optional[List[Tuple[str,
                                               str]]] = None) -> dag_lib.Dag:
    """Loads a chain DAG (pipeline) from a list of YAML configs.

    A pipeline YAML can have an optional header as the first document with:
    - name: The pipeline name
    - execution: Must be 'serial' or omitted (omitted defaults to serial)

    If the first document contains only pipeline header fields, it's treated
    as the header. Otherwise, all documents are treated as task definitions.
    """
    dag_name = None
    first_config = configs[0] if configs else None

    # Check if the first document is a pipeline header.
    # A header has only 'name', or 'name' + 'execution' (for explicit serial).
    is_header = False
    if first_config is not None:
        first_keys = set(first_config.keys())
        is_header = first_keys == {'name'
                                  } or first_keys == {'name', 'execution'}

    if is_header:
        assert first_config is not None  # For mypy
        dag_name = first_config['name']
        # Validate execution mode if specified
        execution = first_config.get('execution')
        if (execution is not None and
                execution != dag_lib.DagExecution.SERIAL.value):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Invalid execution mode for pipeline: {execution!r}. '
                    'Pipelines must use execution: serial (or omit it). '
                    'Use execution: parallel for job groups.')
        configs = configs[1:]
    elif len(configs) == 1 and first_config is not None:
        # Single document: the task itself may have a name
        dag_name = first_config.get('name')

    if not configs:
        # YAML has only header. Still instantiate a task.
        configs = [{'name': dag_name}]

    current_task = None
    with dag_lib.Dag() as dag:
        for task_config in configs:
            if task_config is None:
                continue
            task = task_lib.Task.from_yaml_config(task_config, env_overrides,
                                                  secrets_overrides)
            if current_task is not None:
                current_task >> task  # pylint: disable=pointless-statement
            current_task = task
    dag.name = dag_name
    # Pipelines have serial execution (explicitly or by default)
    dag.execution = dag_lib.DagExecution.SERIAL
    return dag


def load_chain_dag_from_yaml(
    path: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secret_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Loads a chain DAG from a YAML file.

    Has special handling for an initial section in YAML that contains only the
    'name' field, which is the DAG name.

    'env_overrides' is a list of (key, value) pairs that will be used to update
    the task's 'envs' section. If it is a chain dag, the envs will be updated
    for all tasks in the chain.

    'secrets_overrides' is a list of (key, value) pairs that will be used to
    update the task's 'secrets' section. If it is a chain dag, the secrets will
    be updated for all tasks in the chain.

    Returns:
      A chain Dag with 1 or more tasks (an empty entrypoint would create a
      trivial task).
    """
    configs = yaml_utils.read_yaml_all(path)
    return _load_chain_dag(configs, env_overrides, secret_overrides)


def load_chain_dag_from_yaml_str(
    yaml_str: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Loads a chain DAG from a YAML string.

    Has special handling for an initial section in YAML that contains only the
    'name' field, which is the DAG name.

    'env_overrides' is a list of (key, value) pairs that will be used to update
    the task's 'envs' section. If it is a chain dag, the envs will be updated
    for all tasks in the chain.

    'secrets_overrides' is a list of (key, value) pairs that will be used to
    update the task's 'secrets' section. If it is a chain dag, the secrets will
    be updated for all tasks in the chain.

    Returns:
      A chain Dag with 1 or more tasks (an empty entrypoint would create a
      trivial task).
    """
    configs = yaml_utils.read_yaml_all_str(yaml_str)
    return _load_chain_dag(configs, env_overrides, secrets_overrides)


def dump_chain_dag_to_yaml_str(dag: dag_lib.Dag,
                               use_user_specified_yaml: bool = False) -> str:
    """Dumps a chain DAG to a YAML string.

    Args:
        dag: the DAG to dump.
        redact_secrets: whether to redact secrets in the YAML string.

    Returns:
        The YAML string.
    """
    assert dag.is_chain(), dag
    configs = [{'name': dag.name}]
    for task in dag.tasks:
        configs.append(
            task.to_yaml_config(
                use_user_specified_yaml=use_user_specified_yaml))
    return yaml_utils.dump_yaml_str(configs)


def dump_chain_dag_to_yaml(dag: dag_lib.Dag, path: str) -> None:
    """Dumps a chain DAG to a YAML file.

    Args:
        dag: the DAG to dump.
        path: the path to the YAML file.
    """
    dag_str = dump_chain_dag_to_yaml_str(dag)
    with open(path, 'w', encoding='utf-8') as f:
        f.write(dag_str)


def dump_dag_to_yaml_str(dag: dag_lib.Dag,
                         use_user_specified_yaml: bool = False) -> str:
    """Dumps a DAG to a YAML string, auto-detecting the DAG type.

    This function automatically chooses the correct serialization format:
    - For JobGroups: uses multi-document YAML with header + jobs
    - For chain DAGs: uses multi-document YAML with name + tasks

    Args:
        dag: the DAG to dump (can be either a chain DAG or JobGroup).
        use_user_specified_yaml: whether to use user-specified YAML format.

    Returns:
        The YAML string.
    """
    if dag.is_job_group():
        return dump_job_group_to_yaml_str(dag, use_user_specified_yaml)
    else:
        return dump_chain_dag_to_yaml_str(dag, use_user_specified_yaml)


def dump_dag_to_yaml(dag: dag_lib.Dag,
                     path: str,
                     use_user_specified_yaml: bool = False) -> None:
    """Dumps a DAG to a YAML file, auto-detecting the DAG type.

    Args:
        dag: the DAG to dump.
        path: the path to the YAML file.
        use_user_specified_yaml: whether to use user-specified YAML format.
    """
    dag_str = dump_dag_to_yaml_str(dag, use_user_specified_yaml)
    with open(path, 'w', encoding='utf-8') as f:
        f.write(dag_str)


def load_dag_from_yaml_str(
    yaml_str: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Loads a DAG from a YAML string, auto-detecting the type.

    This function automatically detects whether the YAML represents a
    JobGroup or a chain DAG and uses the appropriate loader.

    Args:
        yaml_str: The YAML string to parse.
        env_overrides: Environment variable overrides for all tasks.
        secrets_overrides: Secrets overrides for all tasks.

    Returns:
        A Dag (either a JobGroup or chain DAG).
    """
    if is_job_group_yaml_str(yaml_str):
        return load_job_group_from_yaml_str(yaml_str, env_overrides,
                                            secrets_overrides)
    else:
        return load_chain_dag_from_yaml_str(yaml_str, env_overrides,
                                            secrets_overrides)


def load_dag_from_yaml(
    path: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Loads a DAG from a YAML file, auto-detecting the type.

    This function automatically detects whether the YAML represents a
    JobGroup or a chain DAG and uses the appropriate loader.

    Args:
        path: Path to the YAML file.
        env_overrides: Environment variable overrides for all tasks.
        secrets_overrides: Secrets overrides for all tasks.

    Returns:
        A Dag (either a JobGroup or chain DAG).
    """
    if is_job_group_yaml(path):
        return load_job_group_from_yaml(path, env_overrides, secrets_overrides)
    else:
        return load_chain_dag_from_yaml(path, env_overrides, secrets_overrides)


def maybe_infer_and_fill_dag_and_task_names(dag: dag_lib.Dag) -> None:
    """Infer and fill the dag/task name if it is None.

    This is mostly for display purpose, to make sure the dag/task name is
    readable and meaningful, instead of a random UUID.
    """
    first_task = dag.tasks[0]
    if len(dag.tasks) == 1:
        if dag.name is not None:
            first_task.name = dag.name
        elif first_task.name is not None:
            dag.name = first_task.name

    if dag.name is None:
        dag.name = cluster_utils.generate_cluster_name()

    if len(dag.tasks) == 1:
        if first_task.name is None:
            first_task.name = dag.name
    else:
        for task_id, task in enumerate(dag.tasks):
            if task.name is None:
                task.name = f'{dag.name}-{task_id}'


def fill_default_config_in_dag_for_job_launch(dag: dag_lib.Dag) -> None:
    for task_ in dag.tasks:

        new_resources_list = []
        default_strategy = registry.JOBS_RECOVERY_STRATEGY_REGISTRY.default
        assert default_strategy is not None
        for resources in list(task_.resources):
            original_job_recovery = resources.job_recovery
            job_recovery: Dict[str, Optional[Union[str, int]]] = {
                'strategy': default_strategy
            }
            if isinstance(original_job_recovery, str):
                job_recovery['strategy'] = original_job_recovery
            elif isinstance(original_job_recovery, dict):
                job_recovery.update(original_job_recovery)
                strategy = job_recovery.get('strategy')
                if strategy is None:
                    job_recovery['strategy'] = default_strategy
            change_default_value: Dict[str, Any] = {
                'job_recovery': job_recovery
            }

            new_resources = resources.copy(**change_default_value)
            new_resources_list.append(new_resources)

        job_recovery_strategy = new_resources_list[0].job_recovery
        for resource in new_resources_list:
            if resource.job_recovery != job_recovery_strategy:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('All resources in the task must have'
                                     'the same job recovery strategy.')

        task_.set_resources(type(task_.resources)(new_resources_list))


def is_job_group_yaml(path: str) -> bool:
    """Check if a YAML file defines a JobGroup.

    A JobGroup YAML is a multi-document YAML where the first document
    contains JobGroup header fields like 'execution'.

    Args:
        path: Path to the YAML file.

    Returns:
        True if this is a JobGroup YAML, False otherwise.
    """
    configs = yaml_utils.read_yaml_all(path)
    return _is_job_group_configs(configs)


def is_job_group_yaml_str(yaml_str: str) -> bool:
    """Check if a YAML string defines a JobGroup."""
    configs = yaml_utils.read_yaml_all_str(yaml_str)
    return _is_job_group_configs(configs)


def _is_job_group_configs(configs: List[Dict[str, Any]]) -> bool:
    """Check if configs represent a JobGroup.

    A YAML is a JobGroup if and only if execution is set to 'parallel'.
    If execution is omitted or set to 'serial', it's a pipeline.
    """
    if not configs or len(configs) < 2:
        # JobGroup needs at least header + 1 job
        return False

    header = configs[0]
    if header is None:
        return False

    # Check if execution mode is explicitly set to 'parallel'
    execution = header.get('execution')
    return execution == dag_lib.DagExecution.PARALLEL.value


def load_job_group_from_yaml(
    path: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Load a JobGroup from a multi-document YAML file.

    JobGroup YAML format:
        ---
        name: my-job-group
        execution: parallel
        ---
        name: trainer
        resources:
          accelerators: H100:8
        run: |
          python train.py
        ---
        name: data-processor
        resources:
          accelerators: V100:4
        run: |
          python process.py

    Args:
        path: Path to the JobGroup YAML file.
        env_overrides: Environment variable overrides for all tasks.
        secrets_overrides: Secrets overrides for all tasks.

    Returns:
        A Dag marked as a JobGroup with all jobs as tasks.

    Raises:
        ValueError: If the YAML is not a valid JobGroup format.
    """
    configs = yaml_utils.read_yaml_all(path)
    return _load_job_group(configs, env_overrides, secrets_overrides)


def load_job_group_from_yaml_str(
    yaml_str: str,
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Load a JobGroup from a multi-document YAML string."""
    configs = yaml_utils.read_yaml_all_str(yaml_str)
    return _load_job_group(configs, env_overrides, secrets_overrides)


def _load_job_group(
    configs: List[Dict[str, Any]],
    env_overrides: Optional[List[Tuple[str, str]]] = None,
    secrets_overrides: Optional[List[Tuple[str, str]]] = None,
) -> dag_lib.Dag:
    """Load a JobGroup from parsed YAML configs.

    Args:
        configs: List of YAML document configs. First is header, rest are jobs.
        env_overrides: Environment variable overrides for all tasks.
        secrets_overrides: Secrets overrides for all tasks.

    Returns:
        A Dag marked as a JobGroup.
    """
    if not configs or len(configs) < 2:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('JobGroup YAML must have at least 2 documents: '
                             'header and at least one job definition.')

    # Parse header
    header = configs[0]
    if header is None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('JobGroup header cannot be empty.')

    # Validate header has required fields
    missing_fields = _JOB_GROUP_REQUIRED_HEADER_FIELDS - set(header.keys())
    if missing_fields:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                f'JobGroup header missing required fields: {missing_fields}')

    # Warn about unknown fields in header
    unknown_fields = set(header.keys()) - _JOB_GROUP_HEADER_FIELDS
    if unknown_fields:
        logger.warning(f'Unknown fields in JobGroup header: {unknown_fields}. '
                       'These will be ignored.')

    group_name = header['name']

    # Validate job group name is safe for shell/filesystem use
    # (alphanumeric, hyphens, underscores only)
    if not group_name or not all(c.isalnum() or c in '-_' for c in group_name):
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                f'Invalid job group name: {group_name!r}. '
                'Name must contain only alphanumeric characters, hyphens, '
                'and underscores.')

    execution_str = header.get('execution', dag_lib.DagExecution.PARALLEL.value)

    # Parse execution mode
    try:
        execution = dag_lib.DagExecution(execution_str)
    except ValueError as e:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                f'Invalid execution mode: {execution_str}. '
                f'Valid options: {[e.value for e in dag_lib.DagExecution]}'
            ) from e

    # Parse job definitions
    job_configs = configs[1:]
    if not job_configs:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('JobGroup must have at least one job definition.')

    # Create DAG using context manager pattern (consistent with _load_chain_dag)
    job_names = set()
    with dag_lib.Dag() as dag:
        for i, job_config in enumerate(job_configs):
            if job_config is None:
                continue

            # Ensure each job has a name
            job_name = job_config.get('name')
            if job_name is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Job {i + 1} in JobGroup must have a "name" field.')

            # Validate job name is safe for shell/filesystem use
            if not all(c.isalnum() or c in '-_' for c in job_name):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Invalid job name: {job_name!r}. '
                        'Name must contain only alphanumeric characters, '
                        'hyphens, and underscores.')

            # Check for duplicate job names
            if job_name in job_names:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Duplicate job name in JobGroup: {job_name}')
            job_names.add(job_name)

            # Create task from job config (auto-added to current dag context)
            task = task_lib.Task.from_yaml_config(job_config, env_overrides,
                                                  secrets_overrides)
            task.name = job_name

    # Set DAG execution properties after context manager
    dag.name = group_name
    dag.set_execution(execution)

    # Parse and validate primary_tasks
    primary_tasks = header.get('primary_tasks')
    if primary_tasks is not None:
        if not isinstance(primary_tasks, list):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'primary_tasks must be a list of job names, '
                                 f'got {type(primary_tasks).__name__}')
        # Empty list is treated as "all jobs are primary"
        if primary_tasks:
            for job_name in primary_tasks:
                if not isinstance(job_name, str):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'primary_tasks entries must be strings, '
                            f'got {type(job_name).__name__}')
                if job_name not in job_names:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'primary_tasks references unknown job: '
                            f'{job_name}. Available jobs: {sorted(job_names)}')
            dag.primary_tasks = primary_tasks
        # Empty list means all jobs are primary (same as not setting it)

    # Parse and validate termination_delay
    termination_delay = header.get('termination_delay')
    if termination_delay is not None:
        if isinstance(termination_delay, (str, int)):
            # Simple format: "30s" or 30
            _validate_time_duration(str(termination_delay), 'termination_delay')
            dag.termination_delay = str(termination_delay)
        elif isinstance(termination_delay, dict):
            # Dict format: {"default": "30s", "replay-buffer": "1m"}
            for key, value in termination_delay.items():
                if not isinstance(key, str) or not isinstance(
                        value, (str, int)):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'termination_delay dict must have string keys and '
                            f'string/int values, got {key!r}: {value!r}')
                if key != 'default' and key not in job_names:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'termination_delay references unknown job: '
                            f'{key}. Available jobs: {sorted(job_names)} '
                            f'(or "default")')
                _validate_time_duration(str(value),
                                        f'termination_delay[{key!r}]')
            # Convert all values to strings
            dag.termination_delay = {
                k: str(v) for k, v in termination_delay.items()
            }
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'termination_delay must be a string, int, or dict, '
                    f'got {type(termination_delay).__name__}')

    logger.info(f'Loaded JobGroup "{group_name}" with {len(dag.tasks)} jobs: '
                f'{[t.name for t in dag.tasks]}')

    return dag


def _validate_time_duration(duration: str, field_name: str) -> None:
    """Validate a time duration string.

    Args:
        duration: Time duration string (e.g., '30s', '5m', '1h', or '30')
        field_name: Name of the field for error messages

    Raises:
        ValueError: If the duration is invalid.
    """
    # TIME_PATTERN_SECONDS already enforces non-negative integers with optional
    # unit suffix, so we only need to check the pattern match.
    if not re.match(constants.TIME_PATTERN_SECONDS, duration):
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                f'Invalid time duration for {field_name}: {duration!r}. '
                f'Expected format: NUMBER[s|m|h|d|w] (e.g., "30s", "5m", "1h")')


def dump_job_group_to_yaml_str(dag: dag_lib.Dag,
                               use_user_specified_yaml: bool = False) -> str:
    """Dump a JobGroup DAG to a multi-document YAML string.

    Args:
        dag: The JobGroup DAG to dump.
        use_user_specified_yaml: Whether to use user-specified YAML format.

    Returns:
        Multi-document YAML string.
    """
    assert dag.is_job_group(), 'DAG is not a JobGroup'

    # Build header
    header: Dict[str, Any] = {
        'name': dag.name,
    }
    if dag.execution is not None:
        header['execution'] = dag.execution.value
    if dag.primary_tasks is not None:
        header['primary_tasks'] = dag.primary_tasks
    if dag.termination_delay is not None:
        header['termination_delay'] = dag.termination_delay

    # Build job configs
    configs: List[Dict[str, Any]] = [header]
    for task in dag.tasks:
        job_config = task.to_yaml_config(
            use_user_specified_yaml=use_user_specified_yaml)
        configs.append(job_config)

    return yaml_utils.dump_yaml_str(configs)



"""Task: a coarse-grained stage in an application."""
import collections
import json
import os
import re
from typing import (Any, Callable, Dict, Iterable, List, Optional, Set, Tuple,
                    Union)

import colorama
from pydantic import SecretStr

from sky import clouds
from sky import dag as dag_lib
from sky import exceptions
from sky import resources as resources_lib
from sky import sky_logging
from sky.data import data_utils
from sky.data import storage as storage_lib
from sky.provision import docker_utils
from sky.serve import service_spec
from sky.skylet import constants
from sky.utils import common_utils
from sky.utils import git
from sky.utils import registry
from sky.utils import schemas
from sky.utils import ux_utils
from sky.utils import volume as volume_lib
from sky.utils import yaml_utils

logger = sky_logging.init_logger(__name__)

_VALID_NAME_REGEX = '[a-zA-Z0-9]+(?:[._-]{1,2}[a-zA-Z0-9]+)*'
_VALID_NAME_DESCR = ('ASCII characters and may contain lowercase and'
                     ' uppercase letters, digits, underscores, periods,'
                     ' and dashes. Must start and end with alphanumeric'
                     ' characters. No triple dashes or underscores.')

_RUN_FN_CHECK_FAIL_MSG = (
    'run command generator must take exactly 2 arguments: node_rank (int) and'
    'a list of node ip addresses (List[str]). Got {run_sig}')


def _is_valid_name(name: Optional[str]) -> bool:
    """Checks if the task name is valid.

    Valid is defined as either NoneType or str with ASCII characters which may
    contain lowercase and uppercase letters, digits, underscores, periods,
    and dashes. Must start and end with alphanumeric characters.
    No triple dashes or underscores.

    Examples:
        some_name_here
        some-name-here
        some__name__here
        some--name--here
        some__name--here
        some.name.here
        some-name_he.re
        this---shouldnt--work
        this___shouldnt_work
        _thisshouldntwork
        thisshouldntwork_
    """
    if name is None:
        return True
    return bool(re.fullmatch(_VALID_NAME_REGEX, name))


def _fill_in_env_vars(
    yaml_field: Dict[str, Any],
    task_envs: Dict[str, str],
) -> Dict[str, Any]:
    """Detects env vars in yaml field and fills them with task_envs.

    Use cases of env vars in file_mounts:
    - dst/src paths; e.g.,
        /model_path/llama-${SIZE}b: s3://llama-weights/llama-${SIZE}b
    - storage's name (bucket name)
    - storage's source (local path)

    Use cases of env vars in service:
    - model type; e.g.,
        service:
          readiness_probe:
            path: /v1/chat/completions
            post_data:
              model: $MODEL_NAME
              messages:
                - role: user
                  content: How to print hello world?
              max_tokens: 1

    We simply dump yaml_field into a json string, and replace env vars using
    regex. This should be safe as yaml config has been schema-validated.

    Env vars of the following forms are detected:
        - ${ENV}
        - $ENV
    where <ENV> must appear in task.envs.
    """
    # TODO(zongheng): support ${ENV:-default}?
    yaml_field_str = json.dumps(yaml_field)

    def replace_var(match):
        var_name = match.group(1)
        # If the variable isn't in the dictionary, return it unchanged
        return task_envs.get(var_name, match.group(0))

    # Pattern for valid env var names in bash.
    pattern = r'\$\{?\b([a-zA-Z_][a-zA-Z0-9_]*)\b\}?'
    yaml_field_str = re.sub(pattern, replace_var, yaml_field_str)
    return json.loads(yaml_field_str)


def _check_docker_login_config(task_envs: Dict[str, str],
                               task_secrets: Dict[str, SecretStr]) -> bool:
    """Validates a valid docker login config in task_envs and task_secrets.

    Docker login variables must be specified together either in envs OR secrets,
    not split across both. If any of the docker login env vars is set, all of
    them must be set in the same location.

    Args:
        task_envs: Environment variables
        task_secrets: Secret variables (optional, defaults to empty dict)

    Returns:
        True if there is a valid docker login config.
        False otherwise.
    Raises:
        ValueError: if docker login configuration is invalid.
    """
    if task_secrets is None:
        task_secrets = {}

    all_keys = constants.DOCKER_LOGIN_ENV_VARS
    envs_keys = all_keys & set(task_envs.keys())
    secrets_keys = all_keys & set(task_secrets.keys())

    # Check if any docker variables exist
    if not envs_keys and not secrets_keys:
        return False

    # Check if variables are split across envs and secrets
    if envs_keys and secrets_keys:
        with ux_utils.print_exception_no_traceback():
            raise ValueError(
                'Docker login variables must be specified together either '
                'in envs OR secrets, not split across both. '
                f'Found in envs: {sorted(envs_keys)}, '
                f'Found in secrets: {sorted(secrets_keys)}')

    # Check if all variables are present in the chosen location
    if envs_keys:
        if len(envs_keys) != len(all_keys):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Docker login variables must be specified together '
                    'in envs. '
                    f'Missing from envs: {sorted(all_keys - envs_keys)}')

    if secrets_keys:
        if len(secrets_keys) != len(all_keys):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Docker login variables must be specified together '
                    'in secrets. '
                    f'Missing from secrets: {sorted(all_keys - secrets_keys)}')

    return True


def _with_docker_login_config(
    resources: Union[Set['resources_lib.Resources'],
                     List['resources_lib.Resources']],
    task_envs: Dict[str, str],
    task_secrets: Dict[str, SecretStr],
) -> Union[Set['resources_lib.Resources'], List['resources_lib.Resources']]:
    if not _check_docker_login_config(task_envs, task_secrets):
        return resources
    envs = task_envs.copy()
    for key, value in task_secrets.items():
        envs[key] = value.get_secret_value()
    docker_login_config = docker_utils.DockerLoginConfig.from_env_vars(envs)

    def _add_docker_login_config(resources: 'resources_lib.Resources'):
        docker_image = resources.extract_docker_image()
        if docker_image is None:
            logger.warning(f'{colorama.Fore.YELLOW}Docker login configs '
                           f'{", ".join(constants.DOCKER_LOGIN_ENV_VARS)} '
                           'are provided, but no docker image is specified '
                           'in `image_id`. The login configs will be '
                           f'ignored.{colorama.Style.RESET_ALL}')
            return resources
        # Already checked in extract_docker_image
        assert resources.image_id is not None and len(
            resources.image_id) == 1, resources.image_id
        region = list(resources.image_id.keys())[0]
        return resources.copy(image_id={region: 'docker:' + docker_image},
                              _docker_login_config=docker_login_config)

    new_resources = []
    for r in resources:
        new_resources.append(_add_docker_login_config(r))
    return type(resources)(new_resources)


def _with_docker_username_for_runpod(
    resources: Union[Set['resources_lib.Resources'],
                     List['resources_lib.Resources']],
    task_envs: Dict[str, str],
    task_secrets: Dict[str, SecretStr],
) -> Union[Set['resources_lib.Resources'], List['resources_lib.Resources']]:
    envs = task_envs.copy()
    for key, value in task_secrets.items():
        envs[key] = value.get_secret_value()
    docker_username_for_runpod = envs.get(
        constants.RUNPOD_DOCKER_USERNAME_ENV_VAR)

    # We should not call r.copy() if docker_username_for_runpod is None,
    # to prevent `DummyResources` instance becoming a `Resources` instance.
    if docker_username_for_runpod is None:
        return resources
    return (type(resources)(
        r.copy(_docker_username_for_runpod=docker_username_for_runpod)
        for r in resources))


def get_plaintext_envs_and_secrets(
    envs_and_secrets: Dict[str, Union[str, SecretStr]],) -> Dict[str, str]:
    return {
        k: v.get_secret_value() if isinstance(v, SecretStr) else v
        for k, v in envs_and_secrets.items()
    }


def get_plaintext_secrets(secrets: Dict[str, SecretStr]) -> Dict[str, str]:
    return {k: v.get_secret_value() for k, v in secrets.items()}


class Task:
    """Task: a computation to be run on the cloud."""

    def __init__(
        self,
        name: Optional[str] = None,
        *,
        setup: Optional[Union[str, List[str]]] = None,
        run: Optional[Union[str, List[str]]] = None,
        envs: Optional[Dict[str, str]] = None,
        secrets: Optional[Dict[str, str]] = None,
        workdir: Optional[Union[str, Dict[str, Any]]] = None,
        num_nodes: Optional[int] = None,
        file_mounts: Optional[Dict[str, str]] = None,
        storage_mounts: Optional[Dict[str, storage_lib.Storage]] = None,
        volumes: Optional[Dict[str, Union[str, Dict[str, Any]]]] = None,
        resources: Optional[Union['resources_lib.Resources',
                                  List['resources_lib.Resources'],
                                  Set['resources_lib.Resources']]] = None,
        # Advanced:
        docker_image: Optional[str] = None,
        event_callback: Optional[str] = None,
        blocked_resources: Optional[Iterable['resources_lib.Resources']] = None,
        # Internal use only.
        _file_mounts_mapping: Optional[Dict[str, str]] = None,
        _volume_mounts: Optional[List[volume_lib.VolumeMount]] = None,
        _metadata: Optional[Dict[str, Any]] = None,
        _user_specified_yaml: Optional[str] = None,
    ):
        """Initializes a Task.

        All fields are optional.  ``Task.run`` is the actual program: either a
        shell command to run (str) or a command generator for different nodes
        (lambda; see below).

        Optionally, call ``Task.set_resources()`` to set the resource
        requirements for this task.  If not set, a default CPU-only requirement
        is assumed (the same as ``sky launch``).

        All setters of this class, ``Task.set_*()``, return ``self``, i.e.,
        they are fluent APIs and can be chained together.

        Example:
            .. code-block:: python

                # A Task that will sync up local workdir '.', containing
                # requirements.txt and train.py.
                sky.Task(setup='pip install requirements.txt',
                         run='python train.py',
                         workdir='.')

                # An empty Task for provisioning a cluster.
                task = sky.Task(num_nodes=n).set_resources(...)

                # Chaining setters.
                sky.Task().set_resources(...).set_file_mounts(...)

        Args:
          name: A string name for the Task for display purposes.
          setup: A setup command(s), which will be run before executing the run
            commands ``run``, and executed under ``workdir``.
          run: The actual command for the task. If not None, either a shell
            command(s) (str, list(str)) or a command generator (callable). If
            latter, it must take a node rank and a list of node addresses as
            input and return a shell command (str) (valid to return None for
            some nodes, in which case no commands are run on them). Run
            commands will be run under ``workdir``. Note the command generator
            should be a self-contained lambda.
          envs: A dictionary of environment variables to set before running the
            setup and run commands.
          secrets: A dictionary of secret environment variables to set before
            running the setup and run commands. These will be redacted in logs
            and YAML output.
          workdir: The local working directory or a git repository.
            For a local working directory, this directory will be synced
            to a location on the remote VM(s), and ``setup`` and ``run``
            commands will be run under that location (thus, they can rely on
            relative paths when invoking binaries).
            If a git repository is provided, the repository will be cloned to
            the working directory and the ``setup`` and ``run`` commands will
            be run under the cloned repository.
          num_nodes: The number of nodes to provision for this Task.  If None,
            treated as 1 node.  If > 1, each node will execute its own
            setup/run command, where ``run`` can either be a str, meaning all
            nodes get the same command, or a lambda, with the semantics
            documented above.
          file_mounts: An optional dict of ``{remote_path: (local_path|cloud
            URI)}``, where remote means the VM(s) on which this Task will
            eventually run on, and local means the node from which the task is
            launched.
          storage_mounts: an optional dict of ``{mount_path: sky.Storage
            object}``, where mount_path is the path inside the remote VM(s)
            where the Storage object will be mounted on.
          volumes: A dict of volumes to be mounted for the task. The dict has
            the form of ``{mount_path: volume_name}`` for external persistent
            volumes, or ``{mount_path: volume_config}`` for ephemeral volumes
            where volume_config is a dict with 'size', and optional type,
            labels, and 'config' fields, etc.
          resources: either a sky.Resources, a set of them, or a list of them.
            A set or a list of resources asks the optimizer to "pick the
            best of these resources" to run this task.
          docker_image: (EXPERIMENTAL: Only in effect when LocalDockerBackend
            is used.) The base docker image that this Task will be built on.
            Defaults to 'gpuci/miniforge-cuda:11.4-devel-ubuntu18.04'.
          event_callback: A bash script that will be executed when the task
            changes state.
          blocked_resources: A set of resources that this task cannot run on.
          _file_mounts_mapping: (Internal use only) A dictionary of file mounts
            mapping.
          _volume_mounts: (Internal use only) A list of volume mounts.
          _metadata: (Internal use only) A dictionary of metadata to be added to
            the task.
          _user_specified_yaml: (Internal use only) A string of user-specified
            YAML config.
        """
        self.name = name
        self.storage_mounts: Dict[str, storage_lib.Storage] = {}
        self.storage_plans: Dict[storage_lib.Storage,
                                 storage_lib.StoreType] = {}
        self._envs = envs or {}
        self._secrets = {}
        if secrets is not None:
            self._secrets = {k: SecretStr(v) for k, v in secrets.items()}
        self._volumes = volumes or {}

        # concatenate commands if given as list
        def _concat(commands: Optional[Union[str, List[str]]]) -> Optional[str]:
            if isinstance(commands, list):
                return '\n'.join(commands)
            return commands

        self.run = _concat(run)
        self.setup = _concat(setup)

        # Validate Docker login configuration early if both envs and secrets
        # contain Docker variables
        if self._envs or self._secrets:
            _check_docker_login_config(self._envs, self._secrets)

        self.workdir = workdir
        self.docker_image = (docker_image if docker_image else
                             'gpuci/miniforge-cuda:11.4-devel-ubuntu18.04')
        self.event_callback = event_callback
        # Ignore type error due to a mypy bug.
        # https://github.com/python/mypy/issues/3004
        self._num_nodes = 1
        self.num_nodes = num_nodes  # type: ignore

        self.inputs: Optional[str] = None
        self.outputs: Optional[str] = None
        self.estimated_inputs_size_gigabytes: Optional[float] = None
        self.estimated_outputs_size_gigabytes: Optional[float] = None
        # Default to CPU VM
        self.resources: Union[List['resources_lib.Resources'],
                              Set['resources_lib.Resources']] = {
                                  resources_lib.Resources()
                              }
        self._service: Optional[service_spec.SkyServiceSpec] = None

        # Resources that this task cannot run on.
        self.blocked_resources = blocked_resources

        self.time_estimator_func: Optional[Callable[['resources_lib.Resources'],
                                                    int]] = None
        self.file_mounts: Optional[Dict[str, str]] = None

        # Only set when 'self' is a jobs controller task: 'self.managed_job_dag'
        # is the underlying managed job dag (sky.Dag object).
        self.managed_job_dag: Optional['dag_lib.Dag'] = None

        # Only set when 'self' is a sky serve controller task.
        self.service_name: Optional[str] = None

        # Filled in by the optimizer.  If None, this Task is not planned.
        self.best_resources: Optional['resources_lib.Resources'] = None

        # For internal use only.
        self.file_mounts_mapping: Optional[Dict[str,
                                                str]] = _file_mounts_mapping
        self.volume_mounts: Optional[List[volume_lib.VolumeMount]] = (
            _volume_mounts)

        self._metadata = _metadata if _metadata is not None else {}

        if resources is not None:
            self.set_resources(resources)
        if storage_mounts is not None:
            self.set_storage_mounts(storage_mounts)
        if file_mounts is not None:
            self.set_file_mounts(file_mounts)

        dag = dag_lib.get_current_dag()
        if dag is not None:
            dag.add(self)

        self._user_specified_yaml = _user_specified_yaml

    def validate(self,
                 skip_file_mounts: bool = False,
                 skip_workdir: bool = False):
        """Validate all fields of the task.

        Args:
            skip_file_mounts: Whether to skip validating file mounts.
            skip_workdir: Whether to skip validating workdir.
        """
        self.validate_name()
        self.validate_run()
        if not skip_workdir:
            self.expand_and_validate_workdir()
        if not skip_file_mounts:
            self.expand_and_validate_file_mounts()
        for r in self.resources:
            r.validate()

    def validate_name(self):
        """Validates if the task name is valid."""
        if not _is_valid_name(self.name):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Invalid task name {self.name}. Valid name: '
                                 f'{_VALID_NAME_DESCR}')

    def validate_run(self):
        """Validates if the run command is valid."""
        if self.run is not None and not isinstance(self.run, str):
            with ux_utils.print_exception_no_traceback():
                raise ValueError('run must be a shell script (str). '
                                 f'Got {type(self.run)}')

    def expand_and_validate_file_mounts(self):
        """Expand file_mounts paths to absolute paths and validate them.

        Note: if this function is called on a remote SkyPilot API server,
        it must be after the client side has sync-ed all files to the
        remote server.
        """
        if self.file_mounts is None:
            return
        for target, source in self.file_mounts.items():
            location = f'file_mounts.{target}: {source}'
            self._validate_mount_path(target, location)
            self._validate_path(source, location)
            if data_utils.is_cloud_store_url(target):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'File mount destination paths cannot be cloud storage')
            if not data_utils.is_cloud_store_url(source):
                self.file_mounts[target] = os.path.abspath(
                    os.path.expanduser(source))
                if not os.path.exists(self.file_mounts[target]
                                     ) and not source.startswith('skypilot:'):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'File mount source {source!r} does not exist '
                            'locally. To fix: check if it exists, and correct '
                            'the path.')

    def _validate_mount_path(self, path: str, location: str):
        self._validate_path(path, location)
        # TODO(zhwu): /home/username/sky_workdir as the target path need
        # to be filtered out as well.
        if (path == constants.SKY_REMOTE_WORKDIR and self.workdir is not None):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Cannot use {constants.SKY_REMOTE_WORKDIR!r} as a '
                    'destination path of a file mount, as it will be used '
                    'by the workdir. If uploading a file/folder to the '
                    'workdir is needed, please specify the full path to '
                    'the file/folder.')

    def _validate_path(self, path: str, location: str):
        if path.endswith('/'):
            with ux_utils.print_exception_no_traceback():
                raise ValueError('Mount paths cannot end with a slash '
                                 f'Found: {path} in {location}')

    def expand_and_validate_workdir(self):
        """Expand workdir to absolute path and validate it.

        Note: if this function is called on a remote SkyPilot API server,
        it must be after the client side has sync-ed all files to the
        remote server.
        """
        if self.workdir is None:
            return
        # Only expand the workdir if it is a string
        if isinstance(self.workdir, dict):
            git_ref = self.workdir.get('ref')
            if git_ref is not None:
                self._metadata['git_commit'] = git_ref
            return
        user_workdir = self.workdir
        self.workdir = os.path.abspath(os.path.expanduser(user_workdir))
        if not os.path.isdir(self.workdir):
            # Symlink to a dir is legal (isdir() follows symlinks).
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Workdir must be a valid directory (or '
                    f'a symlink to a directory). {user_workdir} not found.')

        self._metadata['git_commit'] = common_utils.get_git_commit(self.workdir)

    @staticmethod
    def from_yaml_config(
        config: Dict[str, Any],
        env_overrides: Optional[List[Tuple[str, str]]] = None,
        secrets_overrides: Optional[List[Tuple[str, str]]] = None,
    ) -> 'Task':
        user_specified_yaml = config.pop('_user_specified_yaml',
                                         yaml_utils.dump_yaml_str(config))
        # More robust handling for 'envs': explicitly convert keys and values to
        # str, since users may pass '123' as keys/values which will get parsed
        # as int causing validate_schema() to fail.
        envs = config.get('envs')
        if envs is not None and isinstance(envs, dict):
            new_envs: Dict[str, Optional[str]] = {}
            for k, v in envs.items():
                if v is not None:
                    new_envs[str(k)] = str(v)
                else:
                    new_envs[str(k)] = None
            config['envs'] = new_envs

        # More robust handling for 'secrets': explicitly convert keys and values
        # to str, since users may pass '123' as keys/values which will get
        # parsed as int causing validate_schema() to fail.
        secrets = config.get('secrets')
        if secrets is not None and isinstance(secrets, dict):
            new_secrets: Dict[str, Optional[str]] = {}
            for k, v in secrets.items():
                if v is not None:
                    new_secrets[str(k)] = str(v)
                else:
                    new_secrets[str(k)] = None
            config['secrets'] = new_secrets

        common_utils.validate_schema(config, schemas.get_task_schema(),
                                     'Invalid task YAML: ')
        if env_overrides is not None:
            # We must override env vars before constructing the Task, because
            # the Storage object creation is eager and it (its name/source
            # fields) may depend on env vars.
            #
            # FIXME(zongheng): The eagerness / how we construct Task's from
            # entrypoint (YAML, CLI args) should be fixed.
            new_envs = config.get('envs', {})
            new_envs.update(env_overrides)
            config['envs'] = new_envs

        if secrets_overrides is not None:
            # Override secrets vars from CLI.
            new_secrets = config.get('secrets', {})
            new_secrets.update(secrets_overrides)
            config['secrets'] = new_secrets

        for k, v in config.get('envs', {}).items():
            if v is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Environment variable {k!r} is None. Please set a '
                        'value for it in task YAML or with --env flag. '
                        f'To set it to be empty, use an empty string ({k}: "" '
                        f'in task YAML or --env {k}="" in CLI).')

        for k, v in config.get('secrets', {}).items():
            if v is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Secret variable {k!r} is None. Please set a '
                        'value for it in task YAML or with --secret flag. '
                        f'To set it to be empty, use an empty string ({k}: "" '
                        f'in task YAML or --secret {k}="" in CLI).')

        # Fill in any Task.envs into file_mounts (src/dst paths, storage
        # name/source).
        env_vars = config.get('envs', {})
        secrets = config.get('secrets', {})
        env_and_secrets = env_vars.copy()
        env_and_secrets.update(secrets)
        if config.get('file_mounts') is not None:
            config['file_mounts'] = _fill_in_env_vars(config['file_mounts'],
                                                      env_and_secrets)

        # Fill in any Task.envs into service (e.g. MODEL_NAME).
        if config.get('service') is not None:
            config['service'] = _fill_in_env_vars(config['service'],
                                                  env_and_secrets)

        # Fill in any Task.envs into workdir
        if config.get('workdir') is not None:
            config['workdir'] = _fill_in_env_vars(config['workdir'],
                                                  env_and_secrets)

        if config.get('volumes') is not None:
            config['volumes'] = _fill_in_env_vars(config['volumes'],
                                                  env_and_secrets)

        task = Task(
            config.pop('name', None),
            run=config.pop('run', None),
            workdir=config.pop('workdir', None),
            setup=config.pop('setup', None),
            num_nodes=config.pop('num_nodes', None),
            envs=config.pop('envs', None),
            secrets=config.pop('secrets', None),
            volumes=config.pop('volumes', None),
            event_callback=config.pop('event_callback', None),
            _file_mounts_mapping=config.pop('file_mounts_mapping', None),
            _metadata=config.pop('_metadata', None),
            _user_specified_yaml=user_specified_yaml,
        )

        # Create lists to store storage objects inlined in file_mounts.
        # These are retained in dicts in the YAML schema and later parsed to
        # storage objects with the storage/storage_mount objects.
        fm_storages = []
        file_mounts = config.pop('file_mounts', None)
        volumes = []
        if file_mounts is not None:
            copy_mounts = {}
            for dst_path, src in file_mounts.items():
                # Check if it is str path
                if isinstance(src, str):
                    copy_mounts[dst_path] = src
                # If the src is not a str path, it is likely a dict. Try to
                # parse storage object.
                elif isinstance(src, dict):
                    if (src.get('store') ==
                            storage_lib.StoreType.VOLUME.value.lower()):
                        # Build the volumes config for resources.
                        volume_config = {
                            'path': dst_path,
                        }
                        if src.get('name'):
                            volume_config['name'] = src.get('name')
                        persistent = src.get('persistent', False)
                        volume_config['auto_delete'] = not persistent
                        volume_config_detail = src.get('config', {})
                        volume_config.update(volume_config_detail)
                        volumes.append(volume_config)
                        source_path = src.get('source')
                        if source_path:
                            # For volume, copy the source path to the
                            # data directory of the volume mount point.
                            copy_mounts[
                                f'{dst_path.rstrip("/")}/data'] = source_path
                    else:
                        fm_storages.append((dst_path, src))
                else:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Unable to parse file_mount '
                                         f'{dst_path}:{src}')
            task.set_file_mounts(copy_mounts)

        task_storage_mounts: Dict[str, storage_lib.Storage] = {}
        all_storages = fm_storages
        for storage in all_storages:
            mount_path = storage[0]
            assert mount_path, 'Storage mount path cannot be empty.'
            try:
                storage_obj = storage_lib.Storage.from_yaml_config(storage[1])
            except exceptions.StorageSourceError as e:
                # Patch the error message to include the mount path, if included
                e.args = (e.args[0].replace('<destination_path>',
                                            mount_path),) + e.args[1:]
                raise e
            task_storage_mounts[mount_path] = storage_obj
        task.set_storage_mounts(task_storage_mounts)

        if config.get('inputs') is not None:
            inputs_dict = config.pop('inputs')
            assert len(inputs_dict) == 1, 'Only one input is allowed.'
            inputs = list(inputs_dict.keys())[0]
            estimated_size_gigabytes = list(inputs_dict.values())[0]
            # TODO: allow option to say (or detect) no download/egress cost.
            task.set_inputs(inputs=inputs,
                            estimated_size_gigabytes=estimated_size_gigabytes)

        if config.get('outputs') is not None:
            outputs_dict = config.pop('outputs')
            assert len(outputs_dict) == 1, 'Only one output is allowed.'
            outputs = list(outputs_dict.keys())[0]
            estimated_size_gigabytes = list(outputs_dict.values())[0]
            task.set_outputs(outputs=outputs,
                             estimated_size_gigabytes=estimated_size_gigabytes)

        # Handle the top-level config field
        config_override = config.pop('config', None)

        # Store the final config override for use in resource setup
        cluster_config_override = config_override

        # Parse resources field.
        resources_config = config.pop('resources', {})
        if cluster_config_override is not None:
            assert resources_config.get('_cluster_config_overrides') is None, (
                'Cannot set _cluster_config_overrides in both resources and '
                'experimental.config_overrides')
            resources_config[
                '_cluster_config_overrides'] = cluster_config_override
        if volumes:
            resources_config['volumes'] = volumes
        task.set_resources(
            resources_lib.Resources.from_yaml_config(resources_config))

        service = config.pop('service', None)
        pool = config.pop('pool', None)
        if service is not None and pool is not None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Cannot set both service and pool in the same task.')

        if service is not None:
            service = service_spec.SkyServiceSpec.from_yaml_config(service)
            task.set_service(service)
        elif pool is not None:
            # When pool is a dict (from top-level pool: in YAML), wrap it
            # properly The schema expects {'pool': {...}} structure, not
            # {'workers': 1, 'pool': True}
            if isinstance(pool, dict):
                # pool is a dict like {'workers': 1, 'max_workers': 3}
                # Wrap it as {'pool': {'workers': 1, 'max_workers': 3}}
                pool_config_dict = {'pool': pool}
            else:
                # pool is a boolean True (shouldn't happen, but handle it)
                pool_config_dict = {'pool': {}}
            pool_spec = service_spec.SkyServiceSpec.from_yaml_config(
                pool_config_dict)
            task.set_service(pool_spec)

        volume_mounts = config.pop('volume_mounts', None)
        if volume_mounts is not None:
            task.volume_mounts = []
            for vol in volume_mounts:
                common_utils.validate_schema(vol,
                                             schemas.get_volume_mount_schema(),
                                             'Invalid volume mount config: ')
                volume_mount = volume_lib.VolumeMount.from_yaml_config(vol)
                task.volume_mounts.append(volume_mount)

        assert not config, f'Invalid task args: {config.keys()}'
        return task

    @staticmethod
    def from_yaml(yaml_path: str) -> 'Task':
        """Initializes a task from a task YAML.

        Example:
            .. code-block:: python

                task = sky.Task.from_yaml('/path/to/task.yaml')

        Args:
          yaml_path: file path to a valid task yaml file.

        Raises:
          ValueError: if the path gets loaded into a str instead of a dict; or
            if there are any other parsing errors.
        """
        with open(os.path.expanduser(yaml_path), 'r', encoding='utf-8') as f:
            # TODO(zongheng): use
            #  https://github.com/yaml/pyyaml/issues/165#issuecomment-430074049
            # to raise errors on duplicate keys.
            user_specified_yaml = f.read()
            return Task.from_yaml_str(user_specified_yaml)

    @staticmethod
    def from_yaml_str(yaml_str: str) -> 'Task':
        """Initializes a task from a task YAML string.

        Example:
            .. code-block:: python

                task = sky.Task.from_yaml_str('yaml_str')
        """
        config = yaml_utils.safe_load(yaml_str)

        if isinstance(config, str):
            with ux_utils.print_exception_no_traceback():
                raise ValueError('YAML loaded as str, not as dict. '
                                 f'Is it correct? content:\n{yaml_str}')

        if config is None:
            config = {}
        config['_user_specified_yaml'] = yaml_str
        return Task.from_yaml_config(config)

    def resolve_and_validate_volumes(self) -> None:
        """Resolve volumes config to volume mounts and validate them.

        Raises:
            exceptions.VolumeNotFoundError: if any volume is not found.
            exceptions.VolumeTopologyConflictError: if there is conflict in the
              volumes and compute topology.
        """
        # Volumes has been resolved, a typical case is that the API server
        # has resolved the volumes and the dag was then submitted to
        # controllers.
        if self.volume_mounts is not None:
            return None
        if not self._volumes:
            return None
        volume_mounts: List[volume_lib.VolumeMount] = []
        for dst_path, vol in self._volumes.items():
            self._validate_mount_path(dst_path, location='volumes')
            # Shortcut for `dst_path: volume_name` (external persistent volume)
            if isinstance(vol, str):
                volume_mount = volume_lib.VolumeMount.resolve(dst_path, vol)
            elif isinstance(vol, dict):
                # Check if this is an ephemeral volume config or external volume
                # with 'size' field
                if 'size' in vol:
                    # This is an ephemeral volume config
                    volume_mount = (
                        volume_lib.VolumeMount.resolve_ephemeral_config(
                            dst_path, vol))
                elif 'name' in vol:
                    # External volume with 'name' field
                    volume_mount = volume_lib.VolumeMount.resolve(
                        dst_path, vol['name'])
                else:
                    raise ValueError(
                        f'Invalid volume config: {dst_path}: {vol}. '
                        'Either "size" (for ephemeral volume) or "name" '
                        '(for external volume) must be set.')
            else:
                raise ValueError(f'Invalid volume config: {dst_path}: {vol}')
            volume_mounts.append(volume_mount)
        # Disable certain access modes
        disabled_modes = {}
        if self.num_nodes > 1:
            disabled_modes[
                volume_lib.VolumeAccessMode.READ_WRITE_ONCE.value] = (
                    'access mode ReadWriteOnce is not supported for '
                    'multi-node tasks.')
            disabled_modes[
                volume_lib.VolumeAccessMode.READ_WRITE_ONCE_POD.value] = (
                    'access mode ReadWriteOncePod is not supported for '
                    'multi-node tasks.')
        # TODO(aylei): generalize access mode to all volume types
        # Record the required topology and the volume that requires it, e.g.
        # {'cloud': ('volume_name', 'aws')}
        topology: Dict[str, Tuple[str, Optional[str]]] = {
            'cloud': ('', None),
            'region': ('', None),
            'zone': ('', None),
        }
        for vol in volume_mounts:
            # Check access mode
            access_mode = vol.volume_config.config.get('access_mode', '')
            if access_mode in disabled_modes:
                raise ValueError(f'Volume {vol.volume_name} with '
                                 f'{disabled_modes[access_mode]}')
            # Skip ephemeral volumes for topology check
            if vol.is_ephemeral:
                continue
            # Check topology
            for key, (vol_name, previous_req) in topology.items():
                req = getattr(vol.volume_config, key)
                if req is not None:
                    if previous_req is not None and req != previous_req:
                        raise exceptions.VolumeTopologyConflictError(
                            f'Volume {vol.volume_name} can only be attached on '
                            f'{key}:{req}, which conflicts with another volume '
                            f'{vol_name} that requires {key}:{previous_req}. '
                            f'Please use different volumes and retry.')
                    topology[key] = (vol.volume_name, req)
        # Now we have the topology requirements from the intersection of all
        # volumes. Check if there is topology conflict with the resources.
        # Volume must have no conflict with ALL resources even if user
        # specifies 'any_of' resources to ensure no resources will conflict
        # with the volumes during failover.

        for res in self.resources:
            for key, (vol_name, vol_req) in topology.items():
                req = getattr(res, key)
                if (req is not None and vol_req is not None and
                        str(req) != vol_req):
                    raise exceptions.VolumeTopologyConflictError(
                        f'The task requires {key}:{req}, which conflicts with '
                        f'the volume constraint {key}:{vol_req}. Please '
                        f'use different volumes and retry.')
        # No topology conflict, we safely override the topology of resources to
        # satisfy the volume constraints.
        override_params = {}
        for key, (vol_name, vol_req) in topology.items():
            if vol_req is not None:
                if key == 'cloud':
                    override_params[key] = registry.CLOUD_REGISTRY.from_str(
                        vol_req)
                else:
                    override_params[key] = vol_req
        logger.debug(
            f'Override resources with volume constraints: {override_params}')
        self.set_resources_override(override_params)
        self.volume_mounts = volume_mounts

    @property
    def num_nodes(self) -> int:
        return self._num_nodes

    @num_nodes.setter
    def num_nodes(self, num_nodes: Optional[int]) -> None:
        if num_nodes is None:
            num_nodes = 1
        if not isinstance(num_nodes, int) or num_nodes <= 0:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'num_nodes should be a positive int. Got: {num_nodes}')
        self._num_nodes = num_nodes

    @property
    def metadata(self) -> Dict[str, Any]:
        return self._metadata

    @property
    def metadata_json(self) -> str:
        return json.dumps(self._metadata)

    @property
    def envs(self) -> Dict[str, str]:
        return self._envs

    @property
    def secrets(self) -> Dict[str, SecretStr]:
        return self._secrets

    @property
    def volumes(self) -> Dict[str, Union[str, Dict[str, Any]]]:
        return self._volumes

    def set_volumes(self, volumes: Dict[str, Union[str, Dict[str,
                                                             Any]]]) -> None:
        """Sets the volumes for this task.

        Args:
          volumes: a dict of ``{mount_path: volume_name}`` for external
            persistent volumes, or ``{mount_path: volume_config}`` for
            ephemeral volumes.
        """
        self._volumes = volumes

    def update_volumes(self, volumes: Dict[str, Union[str, Dict[str,
                                                                Any]]]) -> None:
        """Updates the volumes for this task."""
        self._volumes.update(volumes)

    def update_envs(
            self, envs: Union[None, List[Tuple[str, str]],
                              Dict[str, str]]) -> 'Task':
        """Updates environment variables for use inside the setup/run commands.

        Args:
          envs: (optional) either a list of ``(env_name, value)`` or a dict
            ``{env_name: value}``.

        Returns:
          self: The current task, with envs updated.

        Raises:
          ValueError: if various invalid inputs errors are detected.
        """
        if envs is None:
            envs = {}
        if isinstance(envs, (list, tuple)):
            keys = set(env[0] for env in envs)
            if len(keys) != len(envs):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('Duplicate env keys provided.')
            envs = dict(envs)
        if isinstance(envs, dict):
            for key in envs:
                if not isinstance(key, str):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError('Env keys must be strings.')
                if not common_utils.is_valid_env_var(key):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Invalid env key: {key}')
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'envs must be List[Tuple[str, str]] or Dict[str, str]: '
                    f'{envs}')
        self._envs.update(envs)
        # If the update_envs() is called after set_resources(), we need to
        # manually update docker login config in task resources, in case the
        # docker login envs are newly added.
        if _check_docker_login_config(self._envs, self._secrets):
            self.resources = _with_docker_login_config(self.resources,
                                                       self._envs,
                                                       self._secrets)
        self.resources = _with_docker_username_for_runpod(
            self.resources, self._envs, self._secrets)
        return self

    def update_secrets(
            self, secrets: Union[None, List[Tuple[str, str]],
                                 Dict[str, str]]) -> 'Task':
        """Updates secret env vars for use inside the setup/run commands.

        Args:
          secrets: (optional) either a list of ``(secret_name, value)`` or a
            dict ``{secret_name: value}``.

        Returns:
          self: The current task, with secrets updated.

        Raises:
          ValueError: if various invalid inputs errors are detected.
        """
        if secrets is None:
            secrets = {}
        if isinstance(secrets, (list, tuple)):
            keys = set(secret[0] for secret in secrets)
            if len(keys) != len(secrets):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('Duplicate secret keys provided.')
            secrets = dict(secrets)
        if isinstance(secrets, dict):
            for key in secrets:
                if not isinstance(key, str):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError('Secret keys must be strings.')
                if not common_utils.is_valid_env_var(key):
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Invalid secret key: {key}')
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'secrets must be List[Tuple[str, str]] or Dict[str, str]: '
                    f'{secrets}')
        for key, value in secrets.items():
            self._secrets[key] = SecretStr(value)
        # Validate Docker login configuration if needed
        if _check_docker_login_config(self._envs, self._secrets):
            self.resources = _with_docker_login_config(self.resources,
                                                       self._envs,
                                                       self._secrets)
        self.resources = _with_docker_username_for_runpod(
            self.resources, self._envs, self._secrets)
        return self

    @property
    def use_spot(self) -> bool:
        return any(r.use_spot for r in self.resources)

    @property
    def envs_and_secrets(self) -> Dict[str, Union[str, SecretStr]]:
        envs = self.envs.copy()
        envs.update(self.secrets)
        return envs

    def set_inputs(self, inputs: str,
                   estimated_size_gigabytes: float) -> 'Task':
        # E.g., 's3://bucket', 'gs://bucket', or None.
        self.inputs = inputs
        self.estimated_inputs_size_gigabytes = estimated_size_gigabytes
        return self

    def get_inputs(self) -> Optional[str]:
        return self.inputs

    def get_estimated_inputs_size_gigabytes(self) -> Optional[float]:
        return self.estimated_inputs_size_gigabytes

    def get_inputs_cloud(self):
        """EXPERIMENTAL: Returns the cloud my inputs live in."""
        assert isinstance(self.inputs, str), self.inputs
        if self.inputs.startswith('s3:'):
            return clouds.AWS()
        elif self.inputs.startswith('gs:'):
            return clouds.GCP()
        elif self.inputs.startswith('cos:'):
            return clouds.IBM()
        else:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'cloud path not supported: {self.inputs}')

    def set_outputs(self, outputs: str,
                    estimated_size_gigabytes: float) -> 'Task':
        self.outputs = outputs
        self.estimated_outputs_size_gigabytes = estimated_size_gigabytes
        return self

    def get_outputs(self) -> Optional[str]:
        return self.outputs

    def get_estimated_outputs_size_gigabytes(self) -> Optional[float]:
        return self.estimated_outputs_size_gigabytes

    @staticmethod
    def _ensure_consistent_priority(
        resources: Union[List['resources_lib.Resources'],
                         Set['resources_lib.Resources']]
    ) -> None:
        priority = None
        for r in resources:
            if r.priority is None:
                continue
            if priority is None:
                priority = r.priority
            else:
                if priority != r.priority:
                    raise ValueError('Priority is not consistent '
                                     f'across resources: {resources}')

    def set_resources(
        self, resources: Union['resources_lib.Resources',
                               List['resources_lib.Resources'],
                               Set['resources_lib.Resources'], Dict[str, Any]]
    ) -> 'Task':
        """Sets the required resources to execute this task.

        If this function is not called for a Task, default resource
        requirements will be used (8 vCPUs).

        Args:
          resources: either a sky.Resources, a set of them, or a list of them.
            A set or a list of resources asks the optimizer to "pick the
            best of these resources" to run this task.
        Returns:
          self: The current task, with resources set.
        """
        if isinstance(resources, dict):
            resources = resources_lib.Resources.from_yaml_config(resources)
        elif isinstance(resources, resources_lib.Resources):
            resources = {resources}
        self._ensure_consistent_priority(resources)
        # TODO(woosuk): Check if the resources are None.
        self.resources = _with_docker_login_config(resources, self.envs,
                                                   self.secrets)
        # Only have effect on RunPod.
        self.resources = _with_docker_username_for_runpod(
            self.resources, self.envs, self.secrets)

        # Evaluate if the task requires FUSE and set the requires_fuse flag
        for _, storage_obj in self.storage_mounts.items():
            if storage_obj.mode in storage_lib.MOUNTABLE_STORAGE_MODES:
                for r in self.resources:
                    r.set_requires_fuse(True)
                break

        return self

    def set_resources_override(self, override_params: Dict[str, Any]) -> 'Task':
        """Sets the override parameters for the resources."""
        new_resources_list = []
        for res in list(self.resources):
            new_resources = res.copy(**override_params)
            new_resources_list.append(new_resources)

        self.set_resources(type(self.resources)(new_resources_list))
        return self

    def get_resource_config(self) -> Dict[str, Any]:
        return _resources_to_config(self.resources,
                                    factor_out_common_fields=True)

    @property
    def service(self) -> Optional[service_spec.SkyServiceSpec]:
        return self._service

    def set_service(self,
                    service: Optional[service_spec.SkyServiceSpec]) -> 'Task':
        """Sets the service spec for this task.

        Args:
          service: a SkyServiceSpec object.

        Returns:
          self: The current task, with service set.
        """
        self._service = service
        return self

    def set_time_estimator(
            self, func: Callable[['resources_lib.Resources'], int]) -> 'Task':
        """Sets a func mapping resources to estimated time (secs).

        This is EXPERIMENTAL.
        """
        self.time_estimator_func = func
        return self

    def estimate_runtime(self, resources):
        """Returns a func mapping resources to estimated time (secs).

        This is EXPERIMENTAL.
        """
        if self.time_estimator_func is None:
            raise NotImplementedError(
                f'Node [{self}] does not have a cost model set; '
                'call set_time_estimator() first')
        return self.time_estimator_func(resources)

    def set_file_mounts(self, file_mounts: Optional[Dict[str, str]]) -> 'Task':
        """Sets the file mounts for this task.

        Useful for syncing datasets, dotfiles, etc.

        File mounts are a dictionary: ``{remote_path: local_path/cloud URI}``.
        Local (or cloud) files/directories will be synced to the specified
        paths on the remote VM(s) where this Task will run.

        Neither source or destimation paths can end with a slash.

        Example:
            .. code-block:: python

                task.set_file_mounts({
                    '~/.dotfile': '/local/.dotfile',
                    # /remote/dir/ will contain the contents of /local/dir/.
                    '/remote/dir': '/local/dir',
                })

        Args:
          file_mounts: an optional dict of ``{remote_path: local_path/cloud
            URI}``, where remote means the VM(s) on which this Task will
            eventually run on, and local means the node from which the task is
            launched.

        Returns:
          self: the current task, with file mounts set.
        """
        self.file_mounts = file_mounts
        return self

    def update_file_mounts(self, file_mounts: Dict[str, str]) -> 'Task':
        """Updates the file mounts for this task.

        Different from set_file_mounts(), this function updates into the
        existing file_mounts (calls ``dict.update()``), rather than
        overwriting it.

        This should be called before provisioning in order to take effect.

        Example:
            .. code-block:: python

                task.update_file_mounts({
                    '~/.config': '~/Documents/config',
                    '/tmp/workdir': '/local/workdir/cnn-cifar10',
                })

        Args:
          file_mounts: a dict of ``{remote_path: local_path/cloud URI}``, where
            remote means the VM(s) on which this Task will eventually run on,
            and local means the node from which the task is launched.

        Returns:
          self: the current task, with file mounts updated.

        Raises:
          ValueError: if input paths are invalid.
        """
        if self.file_mounts is None:
            self.file_mounts = {}
        assert self.file_mounts is not None
        self.file_mounts.update(file_mounts)
        self.expand_and_validate_file_mounts()
        return self

    def set_storage_mounts(
        self,
        storage_mounts: Optional[Dict[str, storage_lib.Storage]],
    ) -> 'Task':
        """Sets the storage mounts for this task.

        Storage mounts are a dictionary: ``{mount_path: sky.Storage object}``,
        each of which mounts a sky.Storage object (a cloud object store bucket)
        to a path inside the remote cluster.

        A sky.Storage object can be created by uploading from a local directory
        (setting ``source``), or backed by an existing cloud bucket (setting
        ``name`` to the bucket name; or setting ``source`` to the bucket URI).

        Example:
            .. code-block:: python

                task.set_storage_mounts({
                    '/remote/imagenet/': sky.Storage(name='my-bucket',
                                                     source='/local/imagenet'),
                })

        Args:
          storage_mounts: an optional dict of ``{mount_path: sky.Storage
            object}``, where mount_path is the path inside the remote VM(s)
            where the Storage object will be mounted on.

        Returns:
          self: The current task, with storage mounts set.

        Raises:
          ValueError: if input paths are invalid.
        """
        if storage_mounts is None:
            self.storage_mounts = {}
            # Clear the requires_fuse flag if no storage mounts are set.
            for r in self.resources:
                r.set_requires_fuse(False)
            return self
        for target, storage_obj in storage_mounts.items():
            # TODO(zhwu): /home/username/sky_workdir as the target path need
            # to be filtered out as well.
            if (target == constants.SKY_REMOTE_WORKDIR and
                    self.workdir is not None):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        f'Cannot use {constants.SKY_REMOTE_WORKDIR!r} as a '
                        'destination path of a file mount, as it will be used '
                        'by the workdir. If uploading a file/folder to the '
                        'workdir is needed, please specify the full path to '
                        'the file/folder.')

            if data_utils.is_cloud_store_url(target):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Storage mount destination path cannot be cloud storage'
                    )

            if storage_obj.mode in storage_lib.MOUNTABLE_STORAGE_MODES:
                # If any storage is using MOUNT mode, we need to enable FUSE in
                # the resources.
                for r in self.resources:
                    r.set_requires_fuse(True)
        # Storage source validation is done in Storage object
        self.storage_mounts = storage_mounts
        return self

    def update_storage_mounts(
            self, storage_mounts: Dict[str, storage_lib.Storage]) -> 'Task':
        """Updates the storage mounts for this task.

        Different from set_storage_mounts(), this function updates into the
        existing storage_mounts (calls ``dict.update()``), rather than
        overwriting it.

        This should be called before provisioning in order to take effect.

        Args:
          storage_mounts: an optional dict of ``{mount_path: sky.Storage
            object}``, where mount_path is the path inside the remote VM(s)
            where the Storage object will be mounted on.

        Returns:
          self: The current task, with storage mounts updated.

        Raises:
          ValueError: if input paths are invalid.
        """
        if not storage_mounts:
            return self
        task_storage_mounts = self.storage_mounts if self.storage_mounts else {}
        task_storage_mounts.update(storage_mounts)
        return self.set_storage_mounts(task_storage_mounts)

    def _get_preferred_store(
            self) -> Tuple[storage_lib.StoreType, Optional[str]]:
        """Returns the preferred store type and region for this task."""
        # TODO(zhwu, romilb): The optimizer should look at the source and
        #  destination to figure out the right stores to use. For now, we
        #  use a heuristic solution to find the store type by the following
        #  order:
        #  1. cloud/region decided in best_resources.
        #  2. cloud/region specified in the task resources.
        #  3. if not specified or the task's cloud does not support storage,
        #     use the first enabled storage cloud with default region.
        # This should be refactored and moved to the optimizer.

        # This check is not needed to support multiple accelerators;
        # We just need to get the storage_cloud.
        # assert len(self.resources) == 1, self.resources
        storage_cloud = None

        enabled_storage_cloud_names = (
            storage_lib.get_cached_enabled_storage_cloud_names_or_refresh(
                raise_if_no_cloud_access=True))

        if self.best_resources is not None:
            storage_cloud = self.best_resources.cloud
            storage_region = self.best_resources.region
        else:
            resources = list(self.resources)[0]
            storage_cloud = resources.cloud
            storage_region = resources.region

        if storage_cloud is not None:
            if str(storage_cloud) not in enabled_storage_cloud_names:
                storage_cloud = None

        storage_cloud_str = None
        if storage_cloud is None:
            storage_cloud_str = enabled_storage_cloud_names[0]
            assert storage_cloud_str is not None, enabled_storage_cloud_names[0]
            storage_region = None  # Use default region in the Store class
        else:
            storage_cloud_str = str(storage_cloud)

        store_type = storage_lib.StoreType.from_cloud(storage_cloud_str)
        return store_type, storage_region

    def sync_storage_mounts(self) -> None:
        """(INTERNAL) Eagerly syncs storage mounts to cloud storage.

        After syncing up, COPY-mode storage mounts are translated into regular
        file_mounts of the form ``{ /remote/path: {s3,gs,..}://<bucket path>
        }``.
        """
        # The same storage can be used multiple times, and we should construct
        # the storage with stores first, so that the storage will be created on
        # the correct cloud.
        name_to_storage = collections.defaultdict(list)
        for storage in self.storage_mounts.values():
            name_to_storage[storage.name].append(storage)
        for storages in name_to_storage.values():
            # Place the storage with most stores first, so that the storage will
            # be created on the correct cloud.
            storage_to_construct = sorted(storages,
                                          key=lambda x: len(x.stores),
                                          reverse=True)
            for storage in storage_to_construct:
                storage.construct()
                assert storage.name is not None, storage
                if not storage.stores:
                    store_type, store_region = self._get_preferred_store()
                    self.storage_plans[storage] = store_type
                    storage.add_store(store_type, store_region)
                else:
                    # We don't need to sync the storage here as if the stores
                    # are not empty, it measn the storage has been synced during
                    # construct() above.
                    # We will download the first store that is added to remote.
                    assert all(store is not None
                               for store in storage.stores.values()), storage
                    self.storage_plans[storage] = list(storage.stores.keys())[0]

        # The following logic converts the storage mounts with COPY mode into
        # inline file mounts with cloud URIs, so that the _execute_file_mounts()
        # in cloud_vm_ray_backend.py can correctly download from the specific
        # cloud storage on the remote cluster.
        # Note that this will cause duplicate destination paths in file_mounts,
        # and storage_mounts, which should be fine as our to_yaml_config() will
        # only dump the storage mount version, i.e. what user specified.
        storage_mounts = self.storage_mounts
        storage_plans = self.storage_plans
        for mnt_path, storage in storage_mounts.items():
            assert storage.name is not None, storage

            if storage.mode == storage_lib.StorageMode.COPY:
                store_type = storage_plans[storage]
                if store_type is storage_lib.StoreType.S3:
                    # TODO: allow for Storage mounting of different clouds
                    if isinstance(storage.source,
                                  str) and storage.source.startswith('s3://'):
                        blob_path = storage.source
                    else:
                        assert storage.name is not None, storage
                        blob_path = 's3://' + storage.name
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.GCS:
                    if isinstance(storage.source,
                                  str) and storage.source.startswith('gs://'):
                        blob_path = storage.source
                    else:
                        assert storage.name is not None, storage
                        blob_path = 'gs://' + storage.name
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.AZURE:
                    if (isinstance(storage.source, str) and
                            data_utils.is_az_container_endpoint(
                                storage.source)):
                        blob_path = storage.source
                    else:
                        assert storage.name is not None, storage
                        store_object = storage.stores[
                            storage_lib.StoreType.AZURE]
                        assert isinstance(store_object,
                                          storage_lib.AzureBlobStore)
                        storage_account_name = store_object.storage_account_name
                        blob_path = data_utils.AZURE_CONTAINER_URL.format(
                            storage_account_name=storage_account_name,
                            container_name=storage.name)
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.R2:
                    if storage.source is not None and not isinstance(
                            storage.source,
                            list) and storage.source.startswith('r2://'):
                        blob_path = storage.source
                    else:
                        blob_path = 'r2://' + storage.name
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.NEBIUS:
                    if storage.source is not None and not isinstance(
                            storage.source,
                            list) and storage.source.startswith('nebius://'):
                        blob_path = storage.source
                    else:
                        blob_path = 'nebius://' + storage.name
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.IBM:
                    if isinstance(storage.source,
                                  str) and storage.source.startswith('cos://'):
                        # source is a cos bucket's uri
                        blob_path = storage.source
                    else:
                        # source is a bucket name.
                        assert storage.name is not None, storage
                        # extract region from rclone.conf
                        cos_region = data_utils.Rclone.get_region_from_rclone(
                            storage.name, data_utils.Rclone.RcloneStores.IBM)
                        blob_path = f'cos://{cos_region}/{storage.name}'
                    blob_path = storage.get_bucket_sub_path_prefix(blob_path)
                    self.update_file_mounts({mnt_path: blob_path})
                elif store_type is storage_lib.StoreType.OCI:
                    if storage.source is not None and not isinstance(
                            storage.source,
                            list) and storage.source.startswith('oci://'):
                        blob_path = storage.source
                    else:
                        blob_path = 'oci://' + storage.name
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                elif store_type is storage_lib.StoreType.COREWEAVE:
                    if storage.source is not None and not isinstance(
                            storage.source,
                            list) and storage.source.startswith('cw://'):
                        blob_path = storage.source
                    else:
                        blob_path = 'cw://' + storage.name
                    self.update_file_mounts({
                        mnt_path: blob_path,
                    })
                else:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(f'Storage Type {store_type} '
                                         'does not exist!')

                # TODO: Delete from storage_mounts, now that the storage is
                # translated into file_mounts. Note: as is, this will break
                # controller_utils.
                # _maybe_translate_local_file_mounts_and_sync_up(), which still
                # needs the storage, but not the file_mounts.

    def get_local_to_remote_file_mounts(self) -> Optional[Dict[str, str]]:
        """Returns file mounts of the form (dst=VM path, src=local path).

        Any cloud object store URIs (gs://, s3://, etc.), either as source or
        destination, are not included.

        INTERNAL: this method is internal-facing.
        """
        if self.file_mounts is None:
            return None
        d = {}
        for k, v in self.file_mounts.items():
            if not data_utils.is_cloud_store_url(
                    k) and not data_utils.is_cloud_store_url(v):
                d[k] = v
        return d

    def is_controller_task(self) -> bool:
        """Returns whether this task is a jobs/serve controller process."""
        return self.managed_job_dag is not None or self.service_name is not None

    def get_cloud_to_remote_file_mounts(self) -> Optional[Dict[str, str]]:
        """Returns file mounts of the form (dst=VM path, src=cloud URL).

        Local-to-remote file mounts are excluded (handled by
        get_local_to_remote_file_mounts()).

        INTERNAL: this method is internal-facing.
        """
        if self.file_mounts is None:
            return None
        d = {}
        for k, v in self.file_mounts.items():
            if not data_utils.is_cloud_store_url(
                    k) and data_utils.is_cloud_store_url(v):
                d[k] = v
        return d

    def update_workdir(self, workdir: Optional[str], git_url: Optional[str],
                       git_ref: Optional[str]) -> 'Task':
        """Updates the task workdir.

        Args:
            workdir: The workdir to update.
            git_url: The git url to update.
            git_ref: The git ref to update.
        """
        if self.workdir is None or isinstance(self.workdir, str):
            if workdir is not None:
                self.workdir = workdir
                return self
            if git_url is not None:
                self.workdir = {}
                self.workdir['url'] = git_url
                if git_ref is not None:
                    self.workdir['ref'] = git_ref
                return self
            return self
        if git_url is not None:
            self.workdir['url'] = git_url
        if git_ref is not None:
            self.workdir['ref'] = git_ref
        return self

    def update_envs_and_secrets_from_workdir(self) -> 'Task':
        """Updates the task envs and secrets from the workdir."""
        if self.workdir is None:
            return self
        if not isinstance(self.workdir, dict):
            return self
        url = self.workdir['url']
        ref = self.workdir.get('ref', '')
        token = os.environ.get(git.GIT_TOKEN_ENV_VAR)
        ssh_key_path = os.environ.get(git.GIT_SSH_KEY_PATH_ENV_VAR)
        try:
            git_repo = git.GitRepo(url, ref, token, ssh_key_path)
            clone_info = git_repo.get_repo_clone_info()
            if clone_info is None:
                return self
            self.envs[git.GIT_URL_ENV_VAR] = clone_info.url
            if ref:
                ref_type = git_repo.get_ref_type()
                if ref_type == git.GitRefType.COMMIT:
                    self.envs[git.GIT_COMMIT_HASH_ENV_VAR] = ref
                elif ref_type == git.GitRefType.BRANCH:
                    self.envs[git.GIT_BRANCH_ENV_VAR] = ref
                elif ref_type == git.GitRefType.TAG:
                    self.envs[git.GIT_TAG_ENV_VAR] = ref
            if clone_info.token is None and clone_info.ssh_key is None:
                return self
            if clone_info.token is not None:
                self.secrets[git.GIT_TOKEN_ENV_VAR] = SecretStr(
                    clone_info.token)
            if clone_info.ssh_key is not None:
                self.secrets[git.GIT_SSH_KEY_ENV_VAR] = SecretStr(
                    clone_info.ssh_key)
        except exceptions.GitError as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'{str(e)}') from None
        return self

    def to_yaml_config(self,
                       use_user_specified_yaml: bool = False) -> Dict[str, Any]:
        """Returns a yaml-style dict representation of the task.

        INTERNAL: this method is internal-facing.
        """
        if use_user_specified_yaml:
            if self._user_specified_yaml is None:
                return self._to_yaml_config(redact_secrets=True)
            config = yaml_utils.safe_load(self._user_specified_yaml)
            if config.get('secrets') is not None:
                config['secrets'] = {k: '<redacted>' for k in config['secrets']}
            return config
        return self._to_yaml_config()

    def _to_yaml_config(self, redact_secrets: bool = False) -> Dict[str, Any]:
        config = {}

        def add_if_not_none(key, value, no_empty: bool = False):
            if no_empty and not value:
                return
            if value is not None:
                config[key] = value

        add_if_not_none('name', self.name)

        tmp_resource_config = _resources_to_config(self.resources)

        add_if_not_none('resources', tmp_resource_config)

        if self.service is not None:
            add_if_not_none('service', self.service.to_yaml_config())

        add_if_not_none('num_nodes', self.num_nodes)

        if self.inputs is not None:
            add_if_not_none('inputs',
                            {self.inputs: self.estimated_inputs_size_gigabytes})
        if self.outputs is not None:
            add_if_not_none(
                'outputs',
                {self.outputs: self.estimated_outputs_size_gigabytes})

        add_if_not_none('setup', self.setup)
        add_if_not_none('workdir', self.workdir)
        add_if_not_none('event_callback', self.event_callback)
        add_if_not_none('run', self.run)

        # Add envs without redaction
        add_if_not_none('envs', self.envs, no_empty=True)

        secrets = self.secrets
        if secrets and not redact_secrets:
            secrets = {k: v.get_secret_value() for k, v in secrets.items()}
        elif secrets and redact_secrets:
            secrets = {k: '<redacted>' for k, v in secrets.items()}
        add_if_not_none('secrets', secrets, no_empty=True)

        add_if_not_none('file_mounts', {})

        if self.file_mounts is not None:
            config['file_mounts'].update(self.file_mounts)

        if self.storage_mounts is not None:
            config['file_mounts'].update({
                mount_path: storage.to_yaml_config()
                for mount_path, storage in self.storage_mounts.items()
            })

        add_if_not_none('file_mounts_mapping', self.file_mounts_mapping)
        add_if_not_none('volumes', self.volumes)
        if self.volume_mounts is not None:
            config['volume_mounts'] = [
                volume_mount.to_yaml_config()
                for volume_mount in self.volume_mounts
            ]
        # we manually check if its empty to not clog up the generated yaml
        add_if_not_none('_metadata', self._metadata if self._metadata else None)
        add_if_not_none('_user_specified_yaml', self._user_specified_yaml)
        return config

    def get_required_cloud_features(
            self) -> Set[clouds.CloudImplementationFeatures]:
        """Returns the required features for this task (but not for resources).

        Features required by the resources are checked separately in
        cloud.get_feasible_launchable_resources().

        INTERNAL: this method is internal-facing.
        """
        required_features = set()

        # Multi-node
        if self.num_nodes > 1:
            required_features.add(clouds.CloudImplementationFeatures.MULTI_NODE)

        # Storage mounting
        for _, storage_mount in self.storage_mounts.items():
            if storage_mount.mode in storage_lib.MOUNTABLE_STORAGE_MODES:
                required_features.add(
                    clouds.CloudImplementationFeatures.STORAGE_MOUNTING)
                break

        return required_features

    def __rshift__(self, b):
        dag = dag_lib.get_current_dag()
        if dag is None:
            raise RuntimeError(
                'Cannot use >> operator outside of a DAG context. '
                'Please use "with sky.Dag() as dag:" to create a DAG context.')
        dag.add_edge(self, b)

    def __repr__(self):
        if isinstance(self.run, str):
            run_msg = self.run.replace('\n', '\\n')
            if len(run_msg) > 20:
                run_msg = f'run=\'{run_msg[:20]}...\''
            else:
                run_msg = f'run=\'{run_msg}\''
        elif self.run is None:
            run_msg = 'run=<empty>'
        else:
            run_msg = 'run=<fn>'

        name_str = ''
        if self.name is not None:
            name_str = f'<name={self.name}>'
        s = f'Task{name_str}({run_msg})'
        if self.inputs is not None:
            s += f'\n  inputs: {self.inputs}'
        if self.outputs is not None:
            s += f'\n  outputs: {self.outputs}'
        if self.num_nodes > 1:
            s += f'\n  nodes: {self.num_nodes}'
        if len(self.resources) > 1:
            resources_str = ('{' + ', '.join(
                r.repr_with_region_zone for r in self.resources) + '}')
            s += f'\n  resources: {resources_str}'
        elif (len(self.resources) == 1 and
              not list(self.resources)[0].is_empty()):
            s += (f'\n  resources: '
                  f'{list(self.resources)[0].repr_with_region_zone}')
        else:
            s += '\n  resources: default instances'
        return s


def _resources_to_config(
        resources: Union[List['resources_lib.Resources'],
                         Set['resources_lib.Resources']],
        factor_out_common_fields: bool = False) -> Dict[str, Any]:
    if len(resources) > 1:
        resource_list: List[Dict[str, Union[str, int]]] = []
        for r in resources:
            resource_list.append(r.to_yaml_config())
        group_key = 'ordered' if isinstance(resources, list) else 'any_of'
        if factor_out_common_fields:
            return _factor_out_common_resource_fields(resource_list, group_key)
        return {group_key: resource_list}
    else:
        return list(resources)[0].to_yaml_config()


def _factor_out_common_resource_fields(configs: List[Dict[str, Union[str,
                                                                     int]]],
                                       group_key: str) -> Dict[str, Any]:
    """Factors out the fields that are common to all resources."""
    return_config: Dict[str, Any] = configs[0].copy()
    if len(configs) > 1:
        for config in configs[1:]:
            for key, value in config.items():
                if key in return_config and return_config[key] != value:
                    del return_config[key]
    num_empty_configs = 0
    for config in configs:
        keys_to_delete = []
        for key, value in config.items():
            if key in return_config:
                keys_to_delete.append(key)
        for key in keys_to_delete:
            del config[key]
        if not config:
            num_empty_configs += 1

    if num_empty_configs == len(configs):
        return return_config
    if len(configs) > 0:
        return_config[group_key] = configs
    return return_config



"""Utilities for processing GPU metrics from Kubernetes clusters."""
import asyncio
import contextlib
import functools
import os
import re
import select
import subprocess
import time
from typing import List, Optional, Tuple

import httpx
import prometheus_client as prom

from sky import sky_logging
from sky.skylet import constants
from sky.utils import common_utils

_SELECT_TIMEOUT = 1
_SELECT_BUFFER_SIZE = 4096

_KB = 2**10
_MB = 2**20
_MEM_BUCKETS = [
    _KB,
    256 * _KB,
    512 * _KB,
    _MB,
    2 * _MB,
    4 * _MB,
    8 * _MB,
    16 * _MB,
    32 * _MB,
    64 * _MB,
    128 * _MB,
    256 * _MB,
    float('inf'),
]

logger = sky_logging.init_logger(__name__)

# Whether the metrics are enabled, cannot be changed at runtime.
METRICS_ENABLED = os.environ.get(constants.ENV_VAR_SERVER_METRICS_ENABLED,
                                 'false').lower() == 'true'

# Time spent processing a piece of code, refer to time_it().
SKY_APISERVER_CODE_DURATION_SECONDS = prom.Histogram(
    'sky_apiserver_code_duration_seconds',
    'Time spent processing code',
    ['name', 'group'],
    buckets=(0.001, 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.25,
             0.35, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.5, 2.75, 3, 3.5, 4, 4.5,
             5, 7.5, 10.0, 12.5, 15.0, 17.5, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0,
             50.0, 55.0, 60.0, 80.0, 120.0, 140.0, 160.0, 180.0, 200.0, 220.0,
             240.0, 260.0, 280.0, 300.0, 320.0, 340.0, 360.0, 380.0, 400.0,
             420.0, 440.0, 460.0, 480.0, 500.0, 520.0, 540.0, 560.0, 580.0,
             600.0, 620.0, 640.0, 660.0, 680.0, 700.0, 720.0, 740.0, 760.0,
             780.0, 800.0, 820.0, 840.0, 860.0, 880.0, 900.0, 920.0, 940.0,
             960.0, 980.0, 1000.0, float('inf')),
)

# Total number of API server requests, grouped by path, method, and status.
SKY_APISERVER_REQUESTS_TOTAL = prom.Counter(
    'sky_apiserver_requests_total',
    'Total number of API server requests',
    ['path', 'method', 'status'],
)

# Time spent processing API server requests, grouped by path, method, and
# status.
SKY_APISERVER_REQUEST_DURATION_SECONDS = prom.Histogram(
    'sky_apiserver_request_duration_seconds',
    'Time spent processing API server requests',
    ['path', 'method', 'status'],
    buckets=(0.001, 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.25,
             0.35, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.5, 2.75, 3, 3.5, 4, 4.5,
             5, 7.5, 10.0, 12.5, 15.0, 17.5, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0,
             50.0, 55.0, 60.0, 80.0, 120.0, 140.0, 160.0, 180.0, 200.0, 220.0,
             240.0, 260.0, 280.0, 300.0, 320.0, 340.0, 360.0, 380.0, 400.0,
             420.0, 440.0, 460.0, 480.0, 500.0, 520.0, 540.0, 560.0, 580.0,
             600.0, 620.0, 640.0, 660.0, 680.0, 700.0, 720.0, 740.0, 760.0,
             780.0, 800.0, 820.0, 840.0, 860.0, 880.0, 900.0, 920.0, 940.0,
             960.0, 980.0, 1000.0, float('inf')),
)

SKY_APISERVER_EVENT_LOOP_LAG_SECONDS = prom.Histogram(
    'sky_apiserver_event_loop_lag_seconds',
    'Scheduling delay of the server event loop',
    ['pid'],
    buckets=(0.001, 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.25,
             0.35, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.5, 2.75, 3, 3.5, 4, 4.5,
             5, 7.5, 10.0, 12.5, 15.0, 17.5, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0,
             50.0, 55.0, 60.0, 80.0, 120.0, 140.0, 160.0, 180.0, 200.0, 220.0,
             240.0, 260.0, 280.0, 300.0, 320.0, 340.0, 360.0, 380.0, 400.0,
             420.0, 440.0, 460.0, 480.0, 500.0, 520.0, 540.0, 560.0, 580.0,
             600.0, 620.0, 640.0, 660.0, 680.0, 700.0, 720.0, 740.0, 760.0,
             780.0, 800.0, 820.0, 840.0, 860.0, 880.0, 900.0, 920.0, 940.0,
             960.0, 980.0, 1000.0, float('inf')),
)

SKY_APISERVER_WEBSOCKET_CONNECTIONS = prom.Gauge(
    'sky_apiserver_websocket_connections',
    'Number of websocket connections',
    ['pid'],
    multiprocess_mode='livesum',
)

SKY_APISERVER_WEBSOCKET_CLOSED_TOTAL = prom.Counter(
    'sky_apiserver_websocket_closed_total',
    'Number of websocket closed',
    ['pid', 'reason'],
)

# The number of execution starts in each worker process, we do not record
# histogram here as the duration has been measured in
# SKY_APISERVER_CODE_DURATION_SECONDS without the worker label (process id).
# Recording histogram WITH worker label will cause high cardinality.
SKY_APISERVER_PROCESS_EXECUTION_START_TOTAL = prom.Counter(
    'sky_apiserver_process_execution_start_total',
    'Total number of execution starts in each worker process',
    ['request', 'pid'],
)

SKY_APISERVER_PROCESS_PEAK_RSS = prom.Gauge(
    'sky_apiserver_process_peak_rss',
    'Peak RSS we saw in each process in last 30 seconds',
    ['pid', 'type'],
)

SKY_APISERVER_PROCESS_CPU_TOTAL = prom.Gauge(
    'sky_apiserver_process_cpu_total',
    'Total CPU times a worker process has been running',
    ['pid', 'type', 'mode'],
)

SKY_APISERVER_REQUEST_MEMORY_USAGE_BYTES = prom.Histogram(
    'sky_apiserver_request_memory_usage_bytes',
    'Peak memory usage of requests', ['name'],
    buckets=_MEM_BUCKETS)

SKY_APISERVER_REQUEST_RSS_INCR_BYTES = prom.Histogram(
    'sky_apiserver_request_rss_incr_bytes',
    'RSS increment after requests', ['name'],
    buckets=_MEM_BUCKETS)

SKY_APISERVER_WEBSOCKET_SSH_LATENCY_SECONDS = prom.Histogram(
    'sky_apiserver_websocket_ssh_latency_seconds',
    ('Time taken for ssh message to go from client to API server and back'
     'to the client. This does not include: latency to reach the pod, '
     'overhead from sending through the k8s port-forward tunnel, or '
     'ssh server lag on the destination pod.'),
    ['pid'],
    buckets=(0.001, 0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.125, 0.15, 0.25,
             0.35, 0.5, 0.75, 1, 1.25, 1.5, 1.75, 2, 2.5, 2.75, 3, 3.5, 4, 4.5,
             5, 7.5, 10.0, 12.5, 15.0, 17.5, 20.0, 25.0, 30.0, 35.0, 40.0, 45.0,
             50.0, 55.0, 60.0, 80.0, 120.0, 140.0, 160.0, 180.0, 200.0, 220.0,
             240.0, 260.0, 280.0, 300.0, 320.0, 340.0, 360.0, 380.0, 400.0,
             420.0, 440.0, 460.0, 480.0, 500.0, 520.0, 540.0, 560.0, 580.0,
             600.0, 620.0, 640.0, 660.0, 680.0, 700.0, 720.0, 740.0, 760.0,
             780.0, 800.0, 820.0, 840.0, 860.0, 880.0, 900.0, 920.0, 940.0,
             960.0, 980.0, 1000.0, float('inf')),
)

SKY_APISERVER_LONG_EXECUTORS = prom.Gauge(
    'sky_apiserver_long_executors',
    'Total number of long-running request executors in the API server',
)

SKY_APISERVER_SHORT_EXECUTORS = prom.Gauge(
    'sky_apiserver_short_executors',
    'Total number of short-running request executors in the API server',
)


@contextlib.contextmanager
def time_it(name: str, group: str = 'default'):
    """Context manager to measure and record code execution duration."""
    if not METRICS_ENABLED:
        yield
    else:
        start_time = time.time()
        try:
            yield
        finally:
            duration = time.time() - start_time
            SKY_APISERVER_CODE_DURATION_SECONDS.labels(
                name=name, group=group).observe(duration)


def time_me(func):
    """Measure the duration of decorated function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not METRICS_ENABLED:
            return func(*args, **kwargs)
        name = f'{func.__module__}/{func.__name__}'
        with time_it(name, group='function'):
            return func(*args, **kwargs)

    return wrapper


def time_me_async(func):
    """Measure the duration of decorated async function."""

    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        if not METRICS_ENABLED:
            return await func(*args, **kwargs)
        name = f'{func.__module__}/{func.__name__}'
        with time_it(name, group='function'):
            return await func(*args, **kwargs)

    return async_wrapper


def start_svc_port_forward(context: str, namespace: str, service: str,
                           service_port: int) -> Tuple[subprocess.Popen, int]:
    """Starts a port forward to a service in a Kubernetes cluster.
    Args:
        context: Kubernetes context name
        namespace: Namespace where the service is located
        service: Service name to port forward to
        service_port: Port on the service to forward to
    Returns:
        Tuple of (subprocess.Popen process, local_port assigned)
    Raises:
        RuntimeError: If port forward fails to start
    """
    start_port_forward_timeout = 10  # 10 second timeout
    terminate_port_forward_timeout = 5  # 5 second timeout

    # Use ':service_port' to let kubectl choose the local port
    cmd = [
        'kubectl', '--context', context, '-n', namespace, 'port-forward',
        f'service/{service}', f':{service_port}'
    ]

    env = os.environ.copy()
    # Use SkyPilot's kubeconfig discovery which respects KUBECONFIG env var
    # (set by credential manager plugin) and falls back to ~/.kube/config.
    # Always set explicitly so subprocess gets the resolved paths even if
    # env var was modified after os.environ was last copied.
    # Import lazily to avoid circular import (metrics -> provision -> clouds
    # -> metrics).
    # pylint: disable=import-outside-toplevel
    from sky.adaptors import kubernetes as kubernetes_adaptors
    from sky.provision.kubernetes import utils as kubernetes_utils
    kubeconfig_paths = kubernetes_utils.get_kubeconfig_paths()
    env['KUBECONFIG'] = kubernetes_adaptors.ENV_KUBECONFIG_PATH_SEPARATOR.join(
        kubeconfig_paths)

    port_forward_process = None
    port_forward_exit = False
    local_port = None
    poller = None
    fd = None

    try:
        # start the port forward process
        port_forward_process = subprocess.Popen(cmd,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.STDOUT,
                                                text=True,
                                                env=env)

        # Use poll() instead of select() to avoid FD_SETSIZE limit
        poller = select.poll()
        assert port_forward_process.stdout is not None
        fd = port_forward_process.stdout.fileno()
        poller.register(fd, select.POLLIN)

        start_time = time.time()
        buffer = ''
        # wait for the port forward to start and extract the local port
        while time.time() - start_time < start_port_forward_timeout:
            if port_forward_process.poll() is not None:
                # port forward process has terminated
                if port_forward_process.returncode != 0:
                    port_forward_exit = True
                break

            # Wait up to 1000ms for data to be available without blocking
            # poll() takes timeout in milliseconds
            events = poller.poll(_SELECT_TIMEOUT * 1000)

            if events:
                # Read available bytes from the FD without blocking
                raw = os.read(fd, _SELECT_BUFFER_SIZE)
                chunk = raw.decode(errors='ignore')
                buffer += chunk
                match = re.search(r'Forwarding from 127\.0\.0\.1:(\d+)', buffer)
                if match:
                    local_port = int(match.group(1))
                    break

            # sleep for 100ms to avoid busy-waiting
            time.sleep(0.1)
    except BaseException:  # pylint: disable=broad-exception-caught
        if port_forward_process:
            stop_svc_port_forward(port_forward_process,
                                  timeout=terminate_port_forward_timeout)
        raise
    finally:
        if poller is not None and fd is not None:
            try:
                poller.unregister(fd)
            except (OSError, ValueError):
                # FD may already be unregistered or invalid
                pass
    if port_forward_exit:
        raise RuntimeError(f'Port forward failed for service {service} in '
                           f'namespace {namespace} on context {context}')
    if local_port is None:
        try:
            if port_forward_process:
                stop_svc_port_forward(port_forward_process,
                                      timeout=terminate_port_forward_timeout)
        finally:
            raise RuntimeError(
                f'Failed to extract local port for service {service} in '
                f'namespace {namespace} on context {context}')

    return port_forward_process, local_port


def stop_svc_port_forward(port_forward_process: subprocess.Popen,
                          timeout: int = 5) -> None:
    """Stops a port forward to a service in a Kubernetes cluster.
    Args:
        port_forward_process: The subprocess.Popen process to terminate
    """
    try:
        port_forward_process.terminate()
        port_forward_process.wait(timeout=timeout)
    except subprocess.TimeoutExpired:
        port_forward_process.kill()
        port_forward_process.wait()


async def send_metrics_request_with_port_forward(
        context: str,
        namespace: str,
        service: str,
        service_port: int,
        endpoint_path: str = '/federate',
        match_patterns: Optional[List[str]] = None,
        timeout: float = 30.0) -> str:
    """Sends a metrics request to a Prometheus endpoint via port forwarding.
    Args:
        context: Kubernetes context name
        namespace: Namespace where the service is located
        service: Service name to port forward to
        service_port: Port on the service to forward to
        endpoint_path: Path to append to the localhost endpoint (e.g.,
            '/federate')
        match_patterns: List of metric patterns to match (for federate
            endpoint)
        timeout: Request timeout in seconds
    Returns:
        Response text containing the metrics
    Raises:
        RuntimeError: If port forward or HTTP request fails
    """
    port_forward_process = None
    try:
        # Start port forward
        port_forward_process, local_port = await asyncio.to_thread(
            start_svc_port_forward, context, namespace, service, service_port)

        # Build endpoint URL
        endpoint = f'http://localhost:{local_port}{endpoint_path}'

        # Make HTTP request
        async with httpx.AsyncClient(timeout=timeout) as client:
            if match_patterns:
                # For federate endpoint, add match[] parameters
                params = [('match[]', pattern) for pattern in match_patterns]
                response = await client.get(endpoint, params=params)
            else:
                response = await client.get(endpoint)

            response.raise_for_status()
            return response.text

    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(f'Failed to send metrics request with port forward: '
                     f'{common_utils.format_exception(e)}')
        raise
    finally:
        # Always clean up port forward
        if port_forward_process:
            await asyncio.to_thread(stop_svc_port_forward, port_forward_process)


async def add_cluster_name_label(metrics_text: str, context: str) -> str:
    """Adds a cluster_name label to each metric line.
    Args:
        metrics_text: The text containing the metrics
        context: The cluster name
    """
    lines = metrics_text.strip().split('\n')
    modified_lines = []

    for line in lines:
        # keep comment lines and empty lines as-is
        if line.startswith('#') or not line.strip():
            modified_lines.append(line)
            continue
        # if line is a metric line with labels, add cluster label
        brace_start = line.find('{')
        brace_end = line.find('}')
        if brace_start != -1 and brace_end != -1:
            metric_name = line[:brace_start]
            existing_labels = line[brace_start + 1:brace_end]
            rest_of_line = line[brace_end + 1:]

            if existing_labels:
                new_labels = f'cluster="{context}",{existing_labels}'
            else:
                new_labels = f'cluster="{context}"'

            modified_line = f'{metric_name}{{{new_labels}}}{rest_of_line}'
            modified_lines.append(modified_line)
        else:
            # keep other lines as-is
            modified_lines.append(line)

    return '\n'.join(modified_lines)


async def get_metrics_for_context(context: str) -> str:
    """Get GPU metrics for a single Kubernetes context.
    Args:
        context: Kubernetes context name
    Returns:
        metrics_text: String containing the metrics
    Raises:
        Exception: If metrics collection fails for any reason
    """
    # Query both DCGM metrics and kube_pod_labels metrics
    # This ensures the dashboard can perform joins to filter by skypilot cluster
    match_patterns = [
        '{__name__=~"node_memory_MemAvailable_bytes|node_memory_MemTotal_bytes|DCGM_.*"}',  # pylint: disable=line-too-long
        'kube_pod_labels',
        'node_cpu_seconds_total{mode="idle"}'
    ]

    # TODO(rohan): don't hardcode the namespace and service name
    metrics_text = await send_metrics_request_with_port_forward(
        context=context,
        namespace='skypilot',
        service='skypilot-prometheus-server',
        service_port=80,
        endpoint_path='/federate',
        match_patterns=match_patterns)

    # add cluster name as a label to each metric line
    metrics_text = await add_cluster_name_label(metrics_text, context)

    return metrics_text



"""Utility functions for deploying local Kubernetes kind clusters."""
import os
import random
import shlex
import subprocess
import tempfile
import textwrap
from typing import Optional, Tuple

from sky import check as sky_check
from sky import sky_logging
from sky.backends import backend_utils
from sky.clouds import cloud as sky_cloud
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.skylet import constants
from sky.skylet import log_lib
from sky.utils import log_utils
from sky.utils import rich_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

# Default path for Kubernetes configuration file
DEFAULT_KUBECONFIG_PATH = os.path.expanduser('~/.kube/config')
DEFAULT_LOCAL_CLUSTER_NAME = 'skypilot'
LOCAL_CLUSTER_PORT_RANGE = 100
LOCAL_CLUSTER_INTERNAL_PORT_START = 30000
LOCAL_CLUSTER_INTERNAL_PORT_END = 30099


def generate_kind_config(port_start: int,
                         num_nodes: int = 1,
                         gpus: bool = False) -> str:
    """Generate a kind cluster config with ports mapped from host to container

    Port range will be [port_start, port_start + LOCAL_CLUSTER_PORT_RANGE)
    Internally, this will map to ports 30000 - 30099

    Args:
        path: Path to generate the config file at
        port_start: Port range start for mappings
        num_nodes: Number of nodes in the cluster
        gpus: If true, initialize kind cluster with GPU support

    Returns:
        The kind cluster config
    """
    internal_start = LOCAL_CLUSTER_INTERNAL_PORT_START
    internal_end = LOCAL_CLUSTER_INTERNAL_PORT_END

    config = textwrap.dedent(f"""
    apiVersion: kind.x-k8s.io/v1alpha4
    kind: Cluster
    kubeadmConfigPatches:
    - |
      kind: ClusterConfiguration
      apiServer:
        extraArgs:
          "service-node-port-range": {internal_start}-{internal_end}
    nodes:
    - role: control-plane
      kubeadmConfigPatches:
      - |
        kind: InitConfiguration
        nodeRegistration:
          kubeletExtraArgs:
            node-labels: "ingress-ready=true"
    """)
    if gpus:
        config += textwrap.indent(
            textwrap.dedent("""
        extraMounts:
        - hostPath: /dev/null
          containerPath: /var/run/nvidia-container-devices/all"""), ' ' * 2)
    config += textwrap.indent(textwrap.dedent("""
      extraPortMappings:"""), ' ' * 2)
    for offset in range(LOCAL_CLUSTER_PORT_RANGE):
        config += textwrap.indent(
            textwrap.dedent(f"""
        - containerPort: {internal_start + offset}
          hostPort: {port_start + offset}
          listenAddress: "0.0.0.0"
          protocol: tcp
        """), ' ' * 2)
    if num_nodes > 1:
        config += '- role: worker\n' * (num_nodes - 1)
    return config


def _get_port_range(name: str, port_start: Optional[int]) -> Tuple[int, int]:
    is_default = name == DEFAULT_LOCAL_CLUSTER_NAME
    if port_start is None:
        if is_default:
            port_start = LOCAL_CLUSTER_INTERNAL_PORT_START
        else:
            port_start = random.randint(301, 399) * 100
    port_end = port_start + LOCAL_CLUSTER_PORT_RANGE - 1

    port_range = f'Current port range: {port_start}-{port_end}'
    if is_default and port_start != LOCAL_CLUSTER_INTERNAL_PORT_START:
        raise ValueError('Default local cluster `skypilot` should have '
                         f'port range from 30000 to 30099. {port_range}.')
    if not is_default and port_start == LOCAL_CLUSTER_INTERNAL_PORT_START:
        raise ValueError('Port range 30000 to 30099 is reserved for '
                         f'default local cluster `skypilot`. {port_range}.')
    if port_start % 100 != 0:
        raise ValueError('Local cluster port start must be a multiple of 100. '
                         f'{port_range}.')

    return port_start, port_end


def deploy_local_cluster(name: Optional[str], port_start: Optional[int],
                         gpus: bool):
    name = name or DEFAULT_LOCAL_CLUSTER_NAME
    port_start, port_end = _get_port_range(name, port_start)
    context_name = f'kind-{name}'
    cluster_created = False

    # Check if GPUs are available on the host
    local_gpus_available = backend_utils.check_local_gpus()
    gpus = gpus and local_gpus_available

    # Check if ~/.kube/config exists:
    if os.path.exists(os.path.expanduser('~/.kube/config')):
        curr_context = kubernetes_utils.get_current_kube_config_context_name()
        if curr_context is not None and curr_context != context_name:
            logger.info(
                f'Current context in kube config: {curr_context}'
                f'\nWill automatically switch to {context_name} after the '
                'local cluster is created.')
    message_str = 'Creating local cluster {}{}...'
    message_str = message_str.format(
        name,
        ' with GPU support (this may take up to 15 minutes)' if gpus else '')

    with tempfile.NamedTemporaryFile(mode='w+', suffix='.yaml',
                                     delete=True) as f:
        # Choose random port range to use on the host machine.
        # Port range is port_start - port_start + 99 (exactly 100 ports).
        logger.debug(f'Using host port range {port_start}-{port_end}')
        f.write(generate_kind_config(port_start, gpus=gpus))
        f.flush()

        path_to_package = os.path.dirname(__file__)
        up_script_path = os.path.join(path_to_package, 'create_cluster.sh')

        # Get directory of script and run it from there
        cwd = os.path.dirname(os.path.abspath(up_script_path))
        run_command = f'{up_script_path} {name} {f.name}'
        if gpus:
            run_command += ' --gpus'
        run_command = shlex.split(run_command)

        # Setup logging paths
        run_timestamp = sky_logging.get_run_timestamp()
        log_path = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp,
                                'local_up.log')
        logger.info(message_str)

        with rich_utils.safe_status(
                ux_utils.spinner_message(message_str,
                                         log_path=log_path,
                                         is_local=True)):
            returncode, _, stderr = log_lib.run_with_log(
                cmd=run_command,
                log_path=log_path,
                require_outputs=True,
                stream_logs=False,
                line_processor=log_utils.SkyLocalUpLineProcessor(
                    log_path=log_path, is_local=True),
                cwd=cwd)

    # Kind always writes to stderr even if it succeeds.
    # If the failure happens after the cluster is created, we need
    # to strip all stderr of "No kind clusters found.", which is
    # printed when querying with kind get clusters.
    stderr = stderr.replace('No kind clusters found.\n', '')

    if returncode == 0:
        cluster_created = True
    elif returncode == 100:
        logger.info(
            ux_utils.finishing_message(
                f'Local cluster {name} already exists.\n',
                log_path=log_path,
                is_local=True,
                follow_up_message=
                'If you want to delete it instead, run: `sky local down --name {name}`'))  # pylint: disable=line-too-long
    else:
        with ux_utils.print_exception_no_traceback():
            log_hint = ux_utils.log_path_hint(log_path, is_local=True)
            raise RuntimeError('Failed to create local cluster. '
                               f'Full log: {log_hint}'
                               f'\nError: {stderr}')
    # Run sky check
    with rich_utils.safe_status('[bold cyan]Running sky check...'):
        sky_check.check_capability(sky_cloud.CloudCapability.COMPUTE,
                                   quiet=True,
                                   clouds=['kubernetes'])
    if cluster_created:
        # Prepare completion message which shows CPU and GPU count
        # Get number of CPUs
        p = subprocess_utils.run(
            'kubectl get nodes -o jsonpath=\'{.items[0].status.capacity.cpu}\'',
            capture_output=True)
        num_cpus = int(p.stdout.decode('utf-8'))

        # GPU count/type parsing
        gpu_message = ''
        gpu_hint = ''
        if gpus:
            # Get GPU model by querying the node labels
            label_name_escaped = 'skypilot.co/accelerator'.replace('.', '\\.')
            gpu_type_cmd = f'kubectl get node {name}-control-plane -o jsonpath=\"{{.metadata.labels[\'{label_name_escaped}\']}}\"'  # pylint: disable=line-too-long
            try:
                # Run the command and capture the output
                gpu_count_output = subprocess.check_output(gpu_type_cmd,
                                                           shell=True,
                                                           text=True)
                gpu_type_str = gpu_count_output.strip() + ' '
            except subprocess.CalledProcessError as e:
                output = str(e.output.decode('utf-8'))
                logger.warning(f'Failed to get GPU type: {output}')
                gpu_type_str = ''

            # Get number of GPUs (sum of nvidia.com/gpu resources)
            gpu_count_command = 'kubectl get nodes -o=jsonpath=\'{range .items[*]}{.status.allocatable.nvidia\\.com/gpu}{\"\\n\"}{end}\' | awk \'{sum += $1} END {print sum}\''  # pylint: disable=line-too-long
            try:
                # Run the command and capture the output
                gpu_count_output = subprocess.check_output(gpu_count_command,
                                                           shell=True,
                                                           text=True)
                gpu_count = gpu_count_output.strip(
                )  # Remove any extra whitespace
                gpu_message = f' and {gpu_count} {gpu_type_str}GPUs'
            except subprocess.CalledProcessError as e:
                output = str(e.output.decode('utf-8'))
                logger.warning(f'Failed to get GPU count: {output}')
                gpu_message = f' with {gpu_type_str}GPU support'

            gpu_hint = (
                '\nHint: To see the list of GPUs in the cluster, '
                'run \'sky gpus list --cloud kubernetes\'') if gpus else ''

        if num_cpus < 2:
            logger.info('Warning: Local cluster has less than 2 CPUs. '
                        'This may cause issues with running tasks.')
        logger.info(
            ux_utils.finishing_message(
                message=(
                    f'Local Kubernetes cluster {name} created successfully '
                    f'with {num_cpus} CPUs{gpu_message} on host port range '
                    f'{port_start}-{port_end}.'),
                log_path=log_path,
                is_local=True,
                follow_up_message=(
                    '\n`sky launch` can now run tasks locally.\n'
                    'Hint: To change the number of CPUs, change your docker '
                    'runtime settings. See https://kind.sigs.k8s.io/docs/user/quick-start/#settings-for-docker-desktop for more info.'  # pylint: disable=line-too-long
                    f'{gpu_hint}')))


def teardown_local_cluster(name: Optional[str] = None):
    name = name or DEFAULT_LOCAL_CLUSTER_NAME
    cluster_removed = False

    path_to_package = os.path.dirname(__file__)
    down_script_path = os.path.join(path_to_package, 'delete_cluster.sh')

    cwd = os.path.dirname(os.path.abspath(down_script_path))
    run_command = f'{down_script_path} {name}'
    run_command = shlex.split(run_command)

    # Setup logging paths
    run_timestamp = sky_logging.get_run_timestamp()
    log_path = os.path.join(constants.SKY_LOGS_DIRECTORY, run_timestamp,
                            'local_down.log')

    with rich_utils.safe_status(
            ux_utils.spinner_message(f'Removing local cluster {name}',
                                     log_path=log_path,
                                     is_local=True)):

        returncode, stdout, stderr = log_lib.run_with_log(cmd=run_command,
                                                          log_path=log_path,
                                                          require_outputs=True,
                                                          stream_logs=False,
                                                          cwd=cwd)
        stderr = stderr.replace('No kind clusters found.\n', '')

        if returncode == 0:
            cluster_removed = True
        elif returncode == 100:
            logger.info(
                ux_utils.error_message(f'Local cluster {name} does not exist.'))
        else:
            with ux_utils.print_exception_no_traceback():
                raise RuntimeError(f'Failed to down local cluster {name}. '
                                   f'Stdout: {stdout}'
                                   f'\nError: {stderr}')
    if cluster_removed:
        # Run sky check
        with rich_utils.safe_status(
                ux_utils.spinner_message('Running sky check...')):
            sky_check.check_capability(sky_cloud.CloudCapability.COMPUTE,
                                       clouds=['kubernetes'],
                                       quiet=True)
        logger.info(
            ux_utils.finishing_message(f'Local cluster {name} removed.',
                                       log_path=log_path,
                                       is_local=True))



"""Networking utilities for JobGroups.

This module provides functions to set up networking between tasks in a JobGroup.

Architecture:
    Layer 1: User Interface (environment variables)
        - SKYPILOT_JOBGROUP_NAME = <job_group_name>

    Layer 2: JobAddressResolver
        - Resolves task addresses for internal networking
        - All tasks run on same infrastructure (cloud + region or K8s cluster)

    Layer 3: NetworkConfigurator
        - Configures network infrastructure (e.g., /etc/hosts injection)
        - Handles platform-specific differences (K8s vs SSH clouds)

Design Goals:
    - Unified interface: All tasks access addresses via environment variables
    - Platform abstraction: K8s uses native DNS, SSH clouds use /etc/hosts
"""
import asyncio
import os
import tempfile
import textwrap
import traceback
import typing
from typing import List, Tuple

from sky import clouds as sky_clouds
from sky import sky_logging
from sky.utils import command_runner

if typing.TYPE_CHECKING:
    from sky import task as task_lib
    from sky.backends import cloud_vm_ray_backend

logger = sky_logging.init_logger(__name__)

# ============================================================================
# Layer 2: JobAddressResolver - Address resolution abstraction
# ============================================================================


def _is_kubernetes(
        handle: 'cloud_vm_ray_backend.CloudVmRayResourceHandle') -> bool:
    """Check if handle is for a Kubernetes cluster."""
    if handle is None:
        return False
    if handle.launched_resources and handle.launched_resources.cloud:
        return handle.launched_resources.cloud.is_same_cloud(
            sky_clouds.Kubernetes())
    return False


def _get_k8s_namespace_from_handle(
        handle: 'cloud_vm_ray_backend.CloudVmRayResourceHandle') -> str:
    """Get Kubernetes namespace from a resource handle.

    Returns:
        Namespace string, defaults to 'default' if not available.
    """
    if handle is None:
        return 'default'

    # Try to get namespace from launched_resources
    if handle.launched_resources and handle.launched_resources.region:
        # In K8s, region is the context name
        try:
            # pylint: disable=import-outside-toplevel
            from sky.provision.kubernetes import utils as k8s_utils
            return k8s_utils.get_kube_config_context_namespace(
                handle.launched_resources.region)
        except Exception as e:  # pylint: disable=broad-except
            logger.debug(f'Failed to get K8s namespace from handle, '
                         f'falling back to default: {e}')

    # Fallback to default namespace
    return 'default'


def _construct_k8s_internal_svc(cluster_name_on_cloud: str, namespace: str,
                                node_idx: int) -> str:
    """Construct Kubernetes internal service DNS URL.

    The pod creation logic guarantees this format.

    Args:
        cluster_name_on_cloud: Cluster name on cloud
        namespace: Kubernetes namespace
        node_idx: Node index (0 for head, 1+ for workers)

    Returns:
        DNS URL like '{cluster}-head.{namespace}.svc.cluster.local'
    """
    if node_idx == 0:
        return f'{cluster_name_on_cloud}-head.{namespace}.svc.cluster.local'
    return (f'{cluster_name_on_cloud}-worker{node_idx}.'
            f'{namespace}.svc.cluster.local')


def _get_job_address(job_name: str,
                     job_group_name: str,
                     node_idx: int = 0) -> str:
    """Get the address for a job node.

    Returns the hostname that will be resolved via /etc/hosts injection.
    Both K8s and SSH clouds use this same hostname format.

    Args:
        job_name: Name of the job.
        job_group_name: Name of the JobGroup.
        node_idx: Node index (0 for head, 1+ for workers). Defaults to 0.

    Returns:
        Hostname string in format: {job_name}-{node_idx}.{job_group_name}
    """
    return f'{job_name}-{node_idx}.{job_group_name}'


# ============================================================================
# Layer 3: NetworkConfigurator - Platform-specific network configuration
# ============================================================================


def _generate_k8s_dns_mappings(
    job_group_name: str,
    tasks_handles: List[Tuple['task_lib.Task',
                              'cloud_vm_ray_backend.CloudVmRayResourceHandle']]
) -> List[Tuple[str, str]]:
    """Generate K8s DNS to hostname mappings for background updater.

    Args:
        job_group_name: Name of the JobGroup.
        tasks_handles: List of (Task, ResourceHandle) tuples.

    Returns:
        List of (k8s_dns, simple_hostname) tuples.
    """
    mappings = []
    for task, handle in tasks_handles:
        if handle is None or not _is_kubernetes(handle):
            continue

        job_name = task.name
        cluster_name_on_cloud = handle.cluster_name_on_cloud
        namespace = _get_k8s_namespace_from_handle(handle)
        num_nodes = (len(handle.stable_internal_external_ips)
                     if handle.stable_internal_external_ips else 1)

        for node_idx in range(num_nodes):
            hostname = f'{job_name}-{node_idx}.{job_group_name}'
            internal_svc = _construct_k8s_internal_svc(cluster_name_on_cloud,
                                                       namespace, node_idx)
            mappings.append((internal_svc, hostname))
            node_type = 'head' if node_idx == 0 else f'worker{node_idx}'
            logger.debug(f'K8s DNS mapping ({node_type}): '
                         f'{internal_svc} -> {hostname}')

    return mappings


def _generate_hosts_entries(
    job_group_name: str,
    tasks_handles: List[Tuple['task_lib.Task',
                              'cloud_vm_ray_backend.CloudVmRayResourceHandle']]
) -> str:
    """Generate /etc/hosts entries for SSH cloud nodes.

    K8s nodes use a background updater to dynamically resolve IPs.

    Args:
        job_group_name: Name of the JobGroup.
        tasks_handles: List of (Task, ResourceHandle) tuples for each task.

    Returns:
        String containing /etc/hosts entries, one per line.
    """
    entries = [f'# JobGroup: {job_group_name}']

    for task, handle in tasks_handles:
        if handle is None:
            logger.warning(f'Skipping task {task.name}: no handle')
            continue

        if _is_kubernetes(handle):
            continue

        if handle.stable_internal_external_ips is None:
            logger.warning(f'Skipping task {task.name}: no IP information')
            continue

        task_name = task.name
        for node_idx, (internal_ip,
                       _) in enumerate(handle.stable_internal_external_ips):
            hostname = f'{task_name}-{node_idx}.{job_group_name}'
            entries.append(f'{internal_ip} {hostname}')
            logger.debug(f'Host entry (SSH): {internal_ip} -> {hostname}')

    return '\n'.join(entries)


async def _inject_hosts_on_node(
    runner: 'command_runner.CommandRunner',
    hosts_content: str,
    job_group_name: str,
) -> bool:
    """Inject /etc/hosts entries on a single node.

    Also creates a marker file to signal that networking setup is complete.

    Args:
        runner: CommandRunner for the target node.
        hosts_content: Content to append to /etc/hosts.
        job_group_name: Name of the JobGroup (for marker file).

    Returns:
        True if successful, False otherwise.
    """
    # pylint: disable=invalid-string-quote
    escaped_content = hosts_content.replace("'", "'\\''")  # noqa: Q000
    marker_file = get_network_ready_marker_path(job_group_name)
    # Use ALIAS_SUDO_TO_EMPTY_FOR_ROOT_CMD to handle containers without sudo
    # but running as root (e.g., pytorch/pytorch images)
    cmd = (
        f'{command_runner.ALIAS_SUDO_TO_EMPTY_FOR_ROOT_CMD} && '
        f"echo '{escaped_content}' | "  # noqa: Q000
        f'sudo tee -a /etc/hosts > /dev/null && touch {marker_file}')
    # pylint: enable=invalid-string-quote

    try:
        loop = asyncio.get_running_loop()
        returncode, _, stderr = await loop.run_in_executor(
            None,
            lambda: runner.run(cmd, stream_logs=False, require_outputs=True))
        if returncode != 0:
            logger.error(f'Failed to inject /etc/hosts: {stderr}')
            return False
        return True
    except Exception as e:  # pylint: disable=broad-except
        logger.error(f'Exception while injecting /etc/hosts: {e}')
        logger.error(traceback.format_exc())
        return False


def generate_k8s_dns_updater_script(dns_mappings: List[Tuple[str, str]],
                                    job_group_name: str) -> str:
    """Generate background script to update /etc/hosts with K8s DNS IPs.

    Args:
        dns_mappings: List of (k8s_dns, simple_hostname) tuples.
        job_group_name: Name of the job group (for process identification).

    Returns:
        Bash script as a string (standalone, without nohup wrapper).
    """
    if not dns_mappings:
        return ''

    mapping_pairs = ' '.join(
        f'{dns}:{hostname}' for dns, hostname in dns_mappings)

    # Note: job_group_name is validated at YAML load time to be shell-safe
    # Use ALIAS_SUDO_TO_EMPTY_FOR_ROOT_CMD to handle containers without sudo
    # but running as root (e.g., pytorch/pytorch images)
    script = textwrap.dedent(f"""\
        #!/bin/bash
        # Background K8s DNS to IP updater for /etc/hosts

        # Disable sudo for root user - handles containers without sudo installed
        {command_runner.ALIAS_SUDO_TO_EMPTY_FOR_ROOT_CMD}

        MAPPINGS="{mapping_pairs}"
        MARKER="# SkyPilot JobGroup K8s entries"

        echo "$(date '+%Y-%m-%d %H:%M:%S') [INFO] Starting DNS updater for {job_group_name}"
        echo "$(date '+%Y-%m-%d %H:%M:%S') [INFO] Monitoring mappings: $MAPPINGS"

        while true; do
          # Build new entries
          new_entries=""
          needs_update=0
          for mapping in $MAPPINGS; do
            k8s_dns="${{mapping%%:*}}"
            simple_name="${{mapping##*:}}"
            # Resolve K8s DNS to IP
            ip=$(getent hosts "$k8s_dns" 2>/dev/null | awk '{{print $1}}')
            if [ -n "$ip" ]; then
              new_entries="${{new_entries}}$ip $simple_name  $MARKER
        "
              # Check if current IP differs from /etc/hosts
              # Note: On first run, current_ip will be empty, triggering update
              current_ip=$(getent hosts "$simple_name" 2>/dev/null | awk '{{print $1}}')
              if [ "$ip" != "$current_ip" ]; then
                needs_update=1
                echo "$(date '+%Y-%m-%d %H:%M:%S') [INFO] IP changed for $simple_name: $current_ip -> $ip"
              fi
            else
              echo "$(date '+%Y-%m-%d %H:%M:%S') [DEBUG] Waiting to resolve $k8s_dns"
            fi
          done

          # Only update /etc/hosts if IPs have changed
          if [ -n "$new_entries" ] && [ $needs_update -eq 1 ]; then
            echo "$(date '+%Y-%m-%d %H:%M:%S') [INFO] Updating /etc/hosts"
            # In K8s, /etc/hosts is mounted by kubelet and cannot be replaced (mv).
            # Instead, we filter and rewrite in-place using tee.
            # 1. Read existing content without our markers
            existing=$(sudo grep -v "$MARKER" /etc/hosts 2>/dev/null || true)
            # 2. Write back existing + new entries using tee
            if echo -e "$existing\\n$new_entries" | sudo tee /etc/hosts > /dev/null; then
              echo "$(date '+%Y-%m-%d %H:%M:%S') [INFO] Successfully updated /etc/hosts"
            else
              echo "$(date '+%Y-%m-%d %H:%M:%S') [ERROR] Failed to update /etc/hosts"
            fi
          fi
          sleep 5
        done
        """)
    return script.strip()


async def _start_k8s_dns_updater_on_node(
    runner: 'command_runner.CommandRunner',
    dns_mappings: List[Tuple[str, str]],
    job_group_name: str,
) -> bool:
    """Start background DNS updater on a K8s node.

    The updater resolves K8s service DNS names to IPs and keeps
    /etc/hosts updated.

    Args:
        runner: CommandRunner for the target node.
        dns_mappings: List of (k8s_dns, simple_hostname) tuples.
        job_group_name: Name of the job group (for process identification).

    Returns:
        True if successful, False otherwise.
    """
    if not dns_mappings:
        return True

    updater_script = generate_k8s_dns_updater_script(dns_mappings,
                                                     job_group_name)

    # Note: job_group_name is validated at YAML load time to be shell-safe
    # (alphanumeric, hyphens, underscores only - see dag_utils.py:477-485)
    # This ensures the process_id is safe for use in pgrep patterns and paths
    process_id = f'skypilot-jobgroup-dns-updater-{job_group_name}'
    script_path = f'/tmp/{process_id}.sh'
    log_path = f'/tmp/{process_id}.log'

    loop = asyncio.get_running_loop()

    try:
        # Upload script via rsync
        with tempfile.NamedTemporaryFile('w',
                                         prefix='sky_dns_updater_',
                                         suffix='.sh',
                                         delete=False) as f:
            f.write(updater_script)
            local_script_path = f.name

        try:
            logger.info(f'Uploading DNS updater script for {job_group_name}...')
            await loop.run_in_executor(
                None, lambda: runner.rsync(source=local_script_path,
                                           target=script_path,
                                           up=True,
                                           stream_logs=False))
            logger.info(f'DNS updater script uploaded to {script_path}')
        finally:
            os.remove(local_script_path)

        # Make executable and run in background, then verify it started.
        # Uses nohup with a subshell to fully detach from kubectl exec.
        # After a brief sleep, pgrep confirms the process is running.
        # Use 0.5s sleep to ensure process is visible on loaded systems.
        # Also create the marker file to signal networking setup is initiated.
        marker_file = get_network_ready_marker_path(job_group_name)
        run_cmd = (f'chmod +x {script_path} && '
                   f'(nohup {script_path} < /dev/null > {log_path} 2>&1 &) && '
                   f'sleep 0.5 && '
                   f'pgrep -f "{process_id}" > /dev/null && '
                   f'touch {marker_file}')
        logger.info(f'Starting DNS updater in background (log: {log_path})...')
        returncode, _, stderr = await loop.run_in_executor(
            None,
            lambda: runner.run(run_cmd, stream_logs=False, require_outputs=True)
        )

        # Exit code 143 (SIGTERM) is expected when kubectl exec closes the
        # connection. The background process continues running despite this.
        if returncode not in (0, 143):
            logger.error(f'Failed to start DNS updater: '
                         f'returncode={returncode}, stderr={stderr}')
            return False

        logger.info('DNS updater started successfully')
        return True
    except Exception as e:  # pylint: disable=broad-except
        logger.error(f'Exception while starting DNS updater: {e}')
        logger.error(traceback.format_exc())
        return False


class NetworkConfigurator:
    """Configures network infrastructure for JobGroups.

    Handles platform-specific network configuration:
    - K8s: No configuration needed (DNS works automatically)
    - SSH clouds: Injects /etc/hosts entries for hostname resolution
    """

    @staticmethod
    async def setup(
        job_group_name: str,
        tasks_handles: List[Tuple[
            'task_lib.Task', 'cloud_vm_ray_backend.CloudVmRayResourceHandle']],
    ) -> bool:
        """Set up network configuration for JobGroup.

        Args:
            job_group_name: Name of the JobGroup.
            tasks_handles: List of (Task, ResourceHandle) tuples.

        Returns:
            True if all configuration succeeded, False otherwise.
        """
        return await NetworkConfigurator._inject_etc_hosts(
            job_group_name, tasks_handles)

    @staticmethod
    async def _inject_etc_hosts(
        job_group_name: str,
        tasks_handles: List[Tuple[
            'task_lib.Task', 'cloud_vm_ray_backend.CloudVmRayResourceHandle']],
    ) -> bool:
        """Inject /etc/hosts entries for all clusters in the JobGroup.

        This maps the unified hostname format to actual addresses:
        - K8s: Write DNS mappings file for skylet's HostUpdater
        - SSH: Inject static internal IPs

        Args:
            job_group_name: Name of the JobGroup.
            tasks_handles: List of (Task, ResourceHandle) tuples for all jobs.

        Returns:
            True if all injections succeeded, False otherwise.
        """
        logger.info(f'Setting up networking on all {len(tasks_handles)} jobs')

        ssh_hosts_content = _generate_hosts_entries(job_group_name,
                                                    tasks_handles)
        k8s_dns_mappings = _generate_k8s_dns_mappings(job_group_name,
                                                      tasks_handles)

        # Each entry: (coroutine, task_name, node_idx, is_k8s)
        setup_tasks: List[Tuple] = []
        for task, handle in tasks_handles:
            if handle is None:
                continue

            is_k8s = _is_kubernetes(handle)
            try:
                runners = handle.get_command_runners()
            except Exception as e:  # pylint: disable=broad-except
                logger.warning(
                    f'Failed to get command runners for {task.name}: {e}')
                continue

            for node_idx, runner in enumerate(runners):
                if is_k8s:
                    coro = _start_k8s_dns_updater_on_node(
                        runner, k8s_dns_mappings, job_group_name)
                    setup_tasks.append((coro, task.name, node_idx, True))
                else:
                    # ssh_hosts_content is always truthy (has header comment)
                    assert ssh_hosts_content, 'unreachable'
                    coro = _inject_hosts_on_node(runner, ssh_hosts_content,
                                                 job_group_name)
                    setup_tasks.append((coro, task.name, node_idx, False))
                logger.debug(
                    f'Queued networking setup for {task.name}-{node_idx}')

        if not setup_tasks:
            logger.warning('No nodes to set up networking')
            return True

        coroutines = [entry[0] for entry in setup_tasks]
        logger.info(f'Setting up networking on {len(coroutines)} nodes...')
        try:
            results = await asyncio.wait_for(asyncio.gather(
                *coroutines, return_exceptions=True),
                                             timeout=60.0)
        except asyncio.TimeoutError:
            logger.error('Networking setup timed out after 60 seconds')
            return False

        success_count = 0
        for i, result in enumerate(results):
            if result is True:
                success_count += 1
                continue

            # Log error details for failed tasks
            _, task_name, node_idx, is_k8s = setup_tasks[i]
            setup_type = 'K8s DNS updater' if is_k8s else '/etc/hosts'
            node_label = f'{task_name}-{node_idx}'

            if isinstance(result, Exception):
                tb_str = ''.join(
                    traceback.format_exception(type(result), result,
                                               result.__traceback__))
                logger.error(
                    f'{setup_type} failed on {node_label}: {result}\n{tb_str}')
            else:
                logger.error(f'{setup_type} failed on {node_label}')

        logger.info(
            f'Hosts injection: {success_count}/{len(results)} succeeded')
        return success_count == len(results)


# ============================================================================
# Layer 4: Public API
# ============================================================================


async def setup_job_group_networking(
    job_group_name: str,
    tasks_handles: List[Tuple['task_lib.Task',
                              'cloud_vm_ray_backend.CloudVmRayResourceHandle']],
) -> bool:
    """Set up networking for all tasks in a JobGroup.

    This is the main entry point for JobGroup networking setup.

    Args:
        job_group_name: Name of the JobGroup.
        tasks_handles: List of (Task, ResourceHandle) tuples for each task.

    Returns:
        True if setup succeeded, False otherwise.
    """
    logger.info(f'Setting up networking for JobGroup: {job_group_name}')
    return await NetworkConfigurator.setup(job_group_name, tasks_handles)


def get_network_ready_marker_path(job_group_name: str) -> str:
    """Get the path to the networking ready marker file.

    This marker file is created by Phase 3 (setup_job_group_networking)
    after /etc/hosts entries are set up. The wait script checks for this
    file before starting the hostname resolution timeout.

    Args:
        job_group_name: Name of the JobGroup.

    Returns:
        Path to the marker file.
    """
    return f'/tmp/skypilot-jobgroup-network-ready-{job_group_name}'


def generate_wait_for_networking_script(job_group_name: str,
                                        other_job_names: List[str]) -> str:
    """Generate a bash script to wait for network setup.

    This script should be prepended to task.setup to ensure networking
    is ready before the task starts.

    The script has two phases:
    1. Wait for the networking ready marker file (created by Phase 3)
    2. Wait for all hostnames to be resolvable

    Args:
        job_group_name: Name of the JobGroup.
        other_job_names: List of other task names in the group to wait for.

    Returns:
        Bash script as a string.
    """
    # Generate hostnames to wait for
    hostnames = [
        f'{task_name}-0.{job_group_name}' for task_name in other_job_names
    ]

    if not hostnames:
        return ''

    hostname_list = ' '.join(hostnames)
    # Note: job_group_name is validated at YAML load time to be shell-safe
    marker_file = get_network_ready_marker_path(job_group_name)
    updater_log = (f'/tmp/skypilot-jobgroup-dns-updater-'
                   f'{job_group_name}.log')
    updater_process = f'skypilot-jobgroup-dns-updater-{job_group_name}'

    # TODO(zhwu): The current handling is not robust against the case where
    # network setup fails. The job will continue but may get stuck if it
    # depends on networking. We should make the job group automatically
    # recover (e.g., re-trigger network setup or restart the job) if the
    # network fails to initialize properly.
    wait_script = textwrap.dedent(f"""
        # Wait for JobGroup networking to be ready (best-effort, non-blocking)
        # If networking fails, we continue anyway to allow job group recovery
        echo "[SkyPilot] Waiting for network setup..."
        NETWORK_READY=true

        # Phase 1: Wait for networking setup to be initiated by controller
        # This marker file is created after Phase 3 sets up /etc/hosts
        MARKER_FILE="{marker_file}"
        MARKER_WAIT=600  # 10 minutes to wait for Phase 3 to start
        MARKER_ELAPSED=0
        echo "[SkyPilot] Waiting for networking initialization marker..."
        while [ ! -f "$MARKER_FILE" ]; do
          if [ $MARKER_ELAPSED -ge $MARKER_WAIT ]; then
            echo "[SkyPilot] Warning: Networking setup not initiated after ${{MARKER_ELAPSED}}s"
            echo "[SkyPilot] Continuing without full network setup (job group may recover later)"
            NETWORK_READY=false
            break
          fi
          if [ $(($MARKER_ELAPSED % 60)) -eq 0 ] && [ $MARKER_ELAPSED -gt 0 ]; then
            echo "[SkyPilot] Still waiting for networking initialization (${{MARKER_ELAPSED}}s elapsed)..."
          fi
          sleep 5
          MARKER_ELAPSED=$((MARKER_ELAPSED + 5))
        done

        if [ "$NETWORK_READY" = "true" ]; then
          echo "[SkyPilot] Networking setup initiated, waiting for hostnames..."

          # Phase 2: Wait for all hostnames to be resolvable
          echo "[SkyPilot] Waiting for hostnames: {hostname_list}"
          HOSTNAMES="{hostname_list}"
          MAX_WAIT=300  # 5 minutes
          ELAPSED=0
          UPDATER_LOG="{updater_log}"
          UPDATER_PROCESS="{updater_process}"
          for hostname in $HOSTNAMES; do
            while ! getent hosts "$hostname" >/dev/null 2>&1; do
              if [ $ELAPSED -ge $MAX_WAIT ]; then
                echo "[SkyPilot] Warning: Network setup timed out for \\"$hostname\\" after ${{ELAPSED}}s"
                echo "[SkyPilot] DNS updater running: $(pgrep -f "$UPDATER_PROCESS" > /dev/null && echo 'yes' || echo 'no')"
                echo "[SkyPilot] Continuing without full network setup (job group may recover later)"
                NETWORK_READY=false
                break 2  # Break out of both loops
              fi
              if [ $(($ELAPSED % 30)) -eq 0 ]; then
                echo "[SkyPilot] Still waiting for $hostname (${{ELAPSED}}s elapsed)..."
              fi
              sleep 2
              ELAPSED=$((ELAPSED + 2))
            done
            if [ "$NETWORK_READY" = "true" ]; then
              echo "[SkyPilot] Hostname $hostname is now resolvable"
            fi
          done
        fi

        if [ "$NETWORK_READY" = "true" ]; then
          echo "[SkyPilot] Network is ready!"
        fi
    """)

    return wait_script.strip()
